import unittest
from quantum_inferno.synth.benchmark_signals import well_tempered_tone

from quantum_inferno.utilities import short_time_fft
from quantum_inferno.utilities import frequency_scales


class MyTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.start_hz = 1
        cls.end_hz = 5
        cls.tukey_alpha = 0.25
        cls.signal, cls.timestamps, cls.fft_nd, cls.sample_rate, cls.freq_center, cls.resolution = well_tempered_tone()
        cls.stft_obj = short_time_fft.get_stft_object_tukey(
            sample_rate_hz=cls.sample_rate,
            tukey_alpha=cls.tukey_alpha,
            segment_length=cls.fft_nd,
            overlap_length=cls.fft_nd // 2,
            scaling="magnitude",
        )
        cls.band_order = 3
        cls.base = 10 ** 0.3
        cls.reference_frequency = 1.0

    def test_get_linear_frequency_bins_range(self):
        result_default = frequency_scales.get_linear_frequency_bins_range(self.sample_rate, self.fft_nd)
        self.assertTrue((result_default == self.stft_obj.f).all())
        result_range = frequency_scales.get_linear_frequency_bins_range(
            self.sample_rate, self.fft_nd, self.start_hz, self.end_hz
        )
        self.assertTrue((result_range == [1.5625, 3.125, 4.6875]).all())

    def test_get_shorttime_fft_frequency_bins(self):
        result = frequency_scales.get_shorttime_fft_frequency_bins(self.sample_rate, self.fft_nd)
        self.assertTrue((result == self.stft_obj.f).all())

    def test_get_band_numbers(self):
        result = frequency_scales.get_band_numbers(
            self.sample_rate, self.band_order, None, None, self.base, self.reference_frequency
        )
        self.assertEqual(len(result), 28)
        result_ranged = frequency_scales.get_band_numbers(
            self.sample_rate, self.band_order, self.start_hz, self.end_hz, self.base, self.reference_frequency
        )
        self.assertEqual(len(result_ranged), 8)

    def test_get_log_central_frequency_bins_range(self):
        result_default = frequency_scales.get_log_central_frequency_bins_range(
            self.sample_rate, self.band_order, None, None, self.base, self.reference_frequency
        )
        self.assertEqual(len(result_default), 28)
        self.assertEqual(result_default[0], 1.0)
        self.assertAlmostEqual(result_default[-1], 501.187233627)
        result_range = frequency_scales.get_log_central_frequency_bins_range(
            self.sample_rate, self.band_order, self.start_hz, self.end_hz, self.base, self.reference_frequency
        )
        self.assertEqual(result_range[0], 1.0)
        self.assertAlmostEqual(result_range[-1], 5.011872336)

    def test_get_log_edge_frequencies(self):
        result_default = frequency_scales.get_log_edge_frequencies(
            self.sample_rate, self.band_order, None, None, self.base, self.reference_frequency
        )
        self.assertEqual(len(result_default), 29)
        self.assertAlmostEqual(result_default[0], 0.891250938)
        self.assertAlmostEqual(result_default[-1], 562.3413251903)
        result_range = frequency_scales.get_log_edge_frequencies(
            self.sample_rate, self.band_order, self.start_hz, self.end_hz, self.base, self.reference_frequency
        )
        self.assertEqual(len(result_range), 9)
        self.assertAlmostEqual(result_range[0], 0.891250938)
        self.assertAlmostEqual(result_range[-1], 5.623413252)



import unittest

import numpy as np

import quantum_inferno.synth.doppler as dplr


class TestDoppler(unittest.TestCase):
    def test_time_duration(self):
        timestamps = np.array([10, 20, 30, 40, 100])
        result = dplr.time_duration(timestamps)
        self.assertEqual(result, 90)



"""
Base templates for plots:
* 3 waveforms
* 1 waveform, 2 mesh
* 1 waveform, 1 mesh
* CW and Power
"""
import math
from typing import List, Optional, Tuple, Union

from matplotlib.collections import QuadMesh
from matplotlib.colorbar import Colorbar
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable, AxesDivider
import numpy as np

import quantum_inferno.utilities.date_time as dt
from quantum_inferno.plot_templates import plot_base as plt_base


def adjust_figure_height(
        figure_size_y: int,
        n_rows: int,
        n_rows_standard: int = 2,
        hspace: float = 0.13
) -> Tuple[float, float, float]:
    """
    Adjust the figure height based on the number of rows to preserve standard panel aspect ratios

    :param figure_size_y: figure height
    :param n_rows: number of rows in figure
    :param n_rows_standard: number of rows in the figure for which height is not adjusted.  Default 2
    :param hspace: height space between panels, fraction of average panel height.  Default 0.13
    :return: adjusted figure height, space param for title, space param for x label
    """
    # space needed for the time label = 10% of the base figure height
    n_px_x_label: float = figure_size_y * 0.1
    # space needed for the title = 6% of the base figure height
    n_px_title: float = figure_size_y * 0.06
    n_px_panel: float = (figure_size_y - n_px_x_label - n_px_title) / ((1. + hspace) * n_rows_standard - hspace)
    n_px_hspace = hspace * n_px_panel
    adjusted_figure_size_y: float = n_px_panel * n_rows + n_px_hspace * (n_rows - 1) + n_px_x_label + n_px_title
    frac_title = 1 - n_px_title / adjusted_figure_size_y
    frac_x_label = n_px_x_label / adjusted_figure_size_y
    return adjusted_figure_size_y, frac_title, frac_x_label


def sanitize_timestamps(time_input: np.ndarray, start_epoch: Optional[float] = None) -> np.ndarray:
    """
    Sanitize timestamps

    :param time_input: array with timestamps
    :param start_epoch: optional start time to sanitize timestamps with.  Default None (use the first timestamp)
    :return: timestamps re-calculated from given epoch_start or first timestamp
    """
    return time_input - (time_input[0] if start_epoch is None else start_epoch)


def get_time_label(
        start_time_epoch: float,
        units_time: str,
        utc_offset_h: float = 0.
) -> str:
    """
    :param start_time_epoch: start time in seconds since epoch UTC
    :param units_time: units of time
    :param utc_offset_h: hours offset from UTC.  Default 0 (UTC time)
    :return: label for time units on a chart
    """
    label: str = f"Time ({units_time})"
    if start_time_epoch != 0:
        start_datetime_epoch = dt.get_datetime_from_timestamp_to_utc(start_time_epoch, utc_offset_h)
        label += f' from UTC {start_datetime_epoch.strftime("%Y-%m-%d %H:%M:%S")}'
    return label


def mesh_time_frequency_edges(
        frequency: np.ndarray,
        time: np.ndarray,
        frequency_ymin: float,
        frequency_ymax: float,
        frequency_scaling: str = "linear"
) -> Tuple[np.ndarray, np.ndarray, float, float]:
    """
    Find time and frequency edges for plotting.  Raises an error if data is invalid.

    :param frequency: frequencies
    :param time: timestamps of the data
    :param frequency_ymin: minimum frequency for y-axis
    :param frequency_ymax: maximum frequency for y-axis
    :param frequency_scaling: "log" or "linear". Default is "linear"
    :return: time and frequency edges, frequency min and max
    """
    if frequency_ymin > frequency_ymax:
        raise ValueError("Highest frequency must be greater than lowest frequency")
    if not np.all(frequency[:-1] <= frequency[1:]):
        raise ValueError("Frequency must be increasing, flip it")
    if not np.all(time[:-1] <= time[1:]):
        raise ValueError("Time must be increasing, flip it")

    t_half_bin: float = np.abs(time[2] - time[1]) / 2.
    t_edge: np.ndarray = np.append(time[0] - t_half_bin, time + t_half_bin)

    if frequency_scaling == "log":
        k_edge: float = np.sqrt(frequency[-1] / frequency[-2])
        f_edge: np.ndarray = np.append(frequency / k_edge, k_edge * frequency[-1])
    else:
        f_half_bin: float = (frequency[2] - frequency[1]) / 2.
        f_edge: np.ndarray = np.append(frequency[0] - f_half_bin, frequency + f_half_bin)

    if frequency_ymin < f_edge[1]:
        frequency_ymin = f_edge[0]
    elif frequency_ymin <= 0 and frequency_scaling == "log":
        frequency_ymin = f_edge[1]
    if frequency_ymax > f_edge[-1]:
        frequency_ymax = f_edge[-1]

    if not isinstance(frequency_ymin, float):
        frequency_ymin = float(frequency_ymin)
    if not isinstance(frequency_ymax, float):
        frequency_ymax = float(frequency_ymax)

    return t_edge, f_edge, frequency_ymin, frequency_ymax


def get_colormesh(axes: plt.Axes, time: np.ndarray, freq: np.ndarray, shading: Union[str, None],
                  mesh_base: plt_base.MeshBase, mesh_panel: plt_base.MeshPanel) -> QuadMesh:
    """
    :param axes: matplotlib axes to invoke the pcolormesh function
    :param time: timestamp array
    :param freq: frequency array
    :param shading: shading value
    :param mesh_base: MeshBase to get info from
    :param mesh_panel: MeshPanel to get info from
    :return: Quadmesh for plotting
    """
    return axes.pcolormesh(time, freq, mesh_panel.tfr, vmin=mesh_panel.color_min, vmax=mesh_panel.color_max,
                           cmap=mesh_base.colormap, shading=shading, snap=True)


def setup_plot(ax: plt.Axes, ylabel_units: str, text_size: int, is_waveform: bool,
               is_bottom: bool, ytick_style: str = "plain"):
    """
    set up a plot with common values
    :param ax: pyplot Axes to set values for
    :param ylabel_units: label for y-axis units
    :param text_size: size of text
    :param is_waveform: if True, the Axes is plotting a waveform, otherwise it's a mesh
    :param is_bottom: if True, the Axes being plotted is the bottom subplot
    :param ytick_style: tick style for waveform y-axis.  Does nothing if is_waveform is False.  Default "plain"
    """
    ax.set_ylabel(ylabel_units, size=text_size)
    ax.tick_params(axis="x", which="both", bottom=is_bottom, labelbottom=is_bottom, labelsize="large")
    ax.tick_params(axis="y", labelsize="large")
    if is_waveform:
        ax.grid(True)
        ax.ticklabel_format(style=ytick_style, scilimits=(0, 0), axis="y")
        ax.yaxis.get_offset_text().set_x(-0.034)


def plot_n_mesh_wf_vert(
        mesh_base: plt_base.MeshBase,
        panels: List[plt_base.MeshPanel],
        wf_base: plt_base.WaveformPlotBase,
        wf_panel: plt_base.WaveformPanel,
        sanitize_times: bool = True,
        use_default_size: bool = True
) -> plt.Figure:
    """
    Plot 1 or more mesh panels above the base waveform panel in a vertical layout.

    :param mesh_base: base values for mesh plots
    :param panels: list of mesh panels to display, in order of display
    :param wf_base: base values for plotting waveforms
    :param wf_panel: WaveformPanel required for figure
    :param sanitize_times: if True, sanitize timestamps.  Default True
    :param use_default_size: if True, use the default size for the plots, otherwise size dynamically.  Default True
    :return: figure to display
    """
    num_panels: int = len(panels) + 1
    time_label: str = get_time_label(wf_base.start_time_epoch, wf_base.units_time)
    # if we need to sanitize times and have a start epoch of 0, use the first timestamp, otherwise use plot base start
    epoch_start = wf_panel.time[0] if wf_base.start_time_epoch == 0 and sanitize_times else wf_base.start_time_epoch
    fig_params = wf_base.params_tfr

    # Time is in the center of the window, frequency is in the fft coefficient center.
    # frequency and time must be increasing!
    t_edge, f_edge, frequency_fix_ymin, frequency_fix_ymax = \
        mesh_time_frequency_edges(frequency=mesh_base.frequency, time=mesh_base.time,
                                  frequency_ymin=mesh_base.frequency_hz_ymin,
                                  frequency_ymax=mesh_base.frequency_hz_ymax,
                                  frequency_scaling=mesh_base.frequency_scaling)

    wf_panel_n_time_zero = sanitize_timestamps(wf_panel.time, epoch_start)
    time_xmin = wf_panel_n_time_zero[0]
    time_xmax = t_edge[-1]

    # pcolormesh must provide corner coordinates, so there will be an offset from step noverlap step size.
    mesh_x, mesh_y, shading = mesh_base.get_colormesh_params()
    if shading is None:
        mesh_x = t_edge
        mesh_y = f_edge

    # format colorbar ticks
    all_cbar_ticks_lens: List[int] = []
    for p in panels:
        all_cbar_ticks_lens.append(max(len(str(math.ceil(p.color_min))), len(str(math.floor(p.color_max)))))
    max_cbar_tick_len: int = sorted(all_cbar_ticks_lens)[-1]
    cbar_tick_fmt: str = f"%-{max_cbar_tick_len}s"

    hspace = 0.13
    if use_default_size:
        title_space = .94
        xlabel_space = .1
        adj_fig_height = fig_params.figure_size_y
    else:
        adj_fig_height, title_space, xlabel_space = adjust_figure_height(fig_params.figure_size_y, num_panels)

    fig, axes = plt.subplots(
        num_panels,
        1,
        figsize=(fig_params.figure_size_x, adj_fig_height),
        sharex=True,
    )

    panel_index = 0
    # main plotting loop
    for p in panels:
        if isinstance(p, plt_base.MeshPanel):
            p.set_color_min_max()
            if not p.is_auto_color_min_max():
                print(f"Mesh panel {panel_index} color scaling with user inputs")
            setup_plot(axes[panel_index], mesh_base.units_frequency, fig_params.text_size, False, False)
            ax_div: AxesDivider = make_axes_locatable(axes[panel_index])
            mesh_panel_cax: plt.Axes = ax_div.append_axes("right", size="1%", pad="0.5%")
            mesh_panel_cbar: Colorbar = fig.colorbar(
                get_colormesh(axes[panel_index], mesh_x, mesh_y, shading, mesh_base, p),
                cax=mesh_panel_cax,
                ticks=[math.ceil(p.color_min), math.floor(p.color_max)],
                format=cbar_tick_fmt)
            mesh_panel_cbar.set_label(p.cbar_units, rotation=270, size=fig_params.text_size)
            axes[panel_index].set_ylim(frequency_fix_ymin, frequency_fix_ymax)
            axes[panel_index].set_yscale(mesh_base.frequency_scaling)
            if mesh_base.frequency_scaling == "linear":
                # Only works for linear range
                axes[panel_index].ticklabel_format(style=p.ytick_style, scilimits=(0, 0), axis="y")
        if panel_index != 0 and panel_index != num_panels - 1:
            axes[panel_index].margins(x=0)
        panel_index += 1

    axes[-1].plot(wf_panel_n_time_zero, wf_panel.sig)
    axes[-1].set_xlim(time_xmin, time_xmax)
    wf_panel.set_y_lims(axes[-1])
    setup_plot(axes[-1], wf_panel.units, fig_params.text_size, True, True, wf_panel.ytick_style)
    ax_div: AxesDivider = make_axes_locatable(axes[-1])
    wf_panel_a_cax: plt.Axes = ax_div.append_axes("right", size="1%", pad="0.5%")
    wf_panel_a_cax.axis("off")

    if wf_base.figure_title_show:
        title = f"{wf_base.figure_title}"
        if wf_base.station_id:
            title += f" at Station {wf_base.station_id}"
        axes[0].set_title(title)
    fig.text(.5, .01, time_label, ha="center", size=fig_params.text_size)
    fig.align_ylabels(axes)
    fig.tight_layout()
    fig.subplots_adjust(bottom=xlabel_space, top=title_space, hspace=hspace)

    return fig


def plot_mesh_wf_vert(
        mesh_base: plt_base.MeshBase,
        mesh_panel: plt_base.MeshPanel,
        wf_base: plt_base.WaveformPlotBase,
        wf_panel: plt_base.WaveformPanel,
        sanitize_times: bool = True,
        use_default_size: bool = True
) -> plt.Figure:
    """
    Specifically plot one mesh and one waveform, vertically

    :param mesh_base: base values for mesh plots
    :param mesh_panel: mesh panel to display
    :param wf_base: base values for plotting waveforms
    :param wf_panel: waveform to display
    :param sanitize_times: if True, sanitize timestamps.  Default True
    :param use_default_size: if True, use the default size for the plots, otherwise size dynamically.  Default True
    :return: Figure to plot
    """
    return plot_n_mesh_wf_vert(mesh_base, [mesh_panel], wf_base, wf_panel, sanitize_times, use_default_size)


def plot_wf_3_vert(
        wf_base: plt_base.WaveformPlotBase,
        wf_panel_a: plt_base.WaveformPanel,
        wf_panel_b: plt_base.WaveformPanel,
        wf_panel_c: plt_base.WaveformPanel,
        sanitize_times: bool = True
) -> plt.Figure:
    """
    plot 3 waveforms

    :param wf_base: base params for plotting waveforms
    :param wf_panel_a: first waveform to plot
    :param wf_panel_b: second waveform to plot
    :param wf_panel_c: third waveform to plot
    :param sanitize_times: if True, sanitize the timestamps.  Default True
    :return: figure to plot
    """
    time_label: str = get_time_label(wf_base.start_time_epoch, wf_base.units_time)

    epoch_start = wf_panel_a.time[0] if wf_base.start_time_epoch == 0 and sanitize_times else wf_base.start_time_epoch
    wf_panel_c_time_zero = sanitize_timestamps(wf_panel_c.time, epoch_start)
    wf_panel_b_time_zero = sanitize_timestamps(wf_panel_b.time, epoch_start)
    wf_panel_a_time_zero = sanitize_timestamps(wf_panel_a.time, epoch_start)

    # Catch cases where there may not be any data
    if wf_panel_a_time_zero[0] == wf_panel_a_time_zero[-1] and wf_panel_b_time_zero[0] == wf_panel_b_time_zero[-1]\
            and wf_panel_c_time_zero[0] == wf_panel_c_time_zero[-1]:
        print("No data to plot for " + wf_base.figure_title)
        return plt.figure()

    if wf_panel_a_time_zero[0] == wf_panel_b_time_zero[0] == wf_panel_c_time_zero[0]:
        time_xmin = wf_panel_a_time_zero[0]
    else:
        time_xmin = np.min([wf_panel_a_time_zero[0], wf_panel_b_time_zero[0], wf_panel_c_time_zero[0]])

    if wf_panel_a_time_zero[-1] == wf_panel_b_time_zero[-1] == wf_panel_c_time_zero[-1]:
        time_xmax = wf_panel_a_time_zero[-1]
    else:
        time_xmax = np.max([wf_panel_a_time_zero[-1], wf_panel_b_time_zero[-1], wf_panel_c_time_zero[-1]])

    # Figure starts here
    fig, axes = plt.subplots(3, 1,
                             figsize=(wf_base.params_tfr.figure_size_x,
                                      wf_base.params_tfr.figure_size_y),
                             sharex=True)
    fig_panel_c: plt.Axes = axes[0]
    axes_iter = 0

    for pnl in ["c", "b", "a"]:
        fig_panel: plt.Axes = axes[axes_iter]
        axes_iter += 1
        # python eval() function allows us to use variables to name other variables
        wf_panel_zero = eval(f"wf_panel_{pnl}_time_zero")
        wf_panel = eval(f"wf_panel_{pnl}")
        fig_panel.plot(wf_panel_zero, wf_panel.sig)
        if wf_base.label_panel_show:
            fig_panel.text(0.01, 0.95, wf_panel.label, transform=fig_panel.transAxes,
                           fontsize=wf_base.params_tfr.text_size,
                           fontweight=wf_base.labels_fontweight, va="top")
        is_bottom = True if pnl == "a" else False
        setup_plot(fig_panel, wf_panel.units, wf_base.params_tfr.text_size, True, is_bottom, "sci")
        fig_panel.set_xlim(time_xmin, time_xmax)

    if wf_base.figure_title_show:
        fig_panel_c.set_title(f"{wf_base.figure_title} at Station {wf_base.station_id}")

    fig.text(.5, .01, time_label, ha="center", size=wf_base.params_tfr.text_size)

    fig.align_ylabels(axes)
    fig.tight_layout()
    fig.subplots_adjust(bottom=.1, hspace=0.13)

    return fig


def setup_cw_power_plot(ax: plt.Axes, y_units: str, x_units: str, text_size: int):
    """
    Set the x and y labels and set tick params for the given Axes

    :param ax: the Axes object to update
    :param y_units: label for y-axis
    :param x_units: label for x-axis
    :param text_size: size of text
    """
    ax.set_ylabel(y_units, size=text_size)
    ax.set_xlabel(f"Time ({x_units})", size=text_size)
    ax.tick_params(axis="x", which="both", bottom=True, labelbottom=True, labelsize="large")
    ax.tick_params(axis="y", which="both", left=True, labelleft=True, labelsize="large")
    ax.grid(True)


def plot_cw_and_power(
        cw_panel: plt_base.CwPanel,
        power_panel: plt_base.PowerPanel,
        cw_plot_base: plt_base.CwPowerPlotBase = plt_base.CwPowerPlotBase()
) -> plt.Figure:
    """
    Template for CW and power plots

    :param cw_panel: CW panel to plot
    :param power_panel: Power panel to plot
    :param cw_plot_base: base parameters for plotting.  Default to default CwPowerPlotBase values
    :return: Figure to plot
    """
    # Catch cases where there may not be any data
    if cw_panel.is_no_data():
        print("No data to plot.")
        return plt.Figure()

    # Figure starts here
    fig, ax = plt.subplots(
        2,
        1,
        figsize=(cw_plot_base.params_tfr.figure_size_x, cw_plot_base.params_tfr.figure_size_y),
    )
    fig_cw_panel: plt.Axes = ax[0]
    fig_power_panel: plt.Axes = ax[1]

    if cw_plot_base.figure_title_show:
        fig_cw_panel.set_title(cw_panel.title, size=cw_plot_base.params_tfr.text_size)
        fig_power_panel.set_title(power_panel.title, size=cw_plot_base.params_tfr.text_size)

    fig_cw_panel.plot(cw_panel.time, cw_panel.sig)
    setup_cw_power_plot(fig_cw_panel, cw_panel.y_units, cw_panel.x_units, cw_plot_base.params_tfr.text_size)

    for i in power_panel.panel_data:
        fig_power_panel.semilogx(i.freq, i.sig, ls=i.linestyle, lw=i.linewidth, label=i.sig_label)
    setup_cw_power_plot(fig_power_panel, power_panel.y_units, power_panel.x_units, cw_plot_base.params_tfr.text_size)
    fig_power_panel.legend()

    fig.tight_layout()
    fig.subplots_adjust()
    return fig



"""
Base classes, constants, and functions used to create plots
"""
from dataclasses import dataclass
from typing import cast, List, Literal, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np

from quantum_inferno.plot_templates.figure_attributes import AudioParams

# sets the default y limits for plots
DEFAULT_YLIM_MIN = -10
DEFAULT_YLIM_MAX = 10

# todo: all enumerated values for plots
WF_Y_SCALING_VALS = ["auto", "symmetric", "positive", "else"]
MESH_SHADING_VALS = ["auto", "gouraud", "flat", "nearest"]  # pyplot pcolormesh.shading
COLORMAP_SCALING_VALS = ["auto", "range", "else"]
YTICK_STYLE_VALS = ["sci", "scientific", "plain"]  # pyplot ytick style
AXIS_SCALE_VALS = ["function", "linear", "log", "functionlog", "symlog", "logit", "asinh"]  # pyplot yscale and xscale


@dataclass
class PlotBase:
    """
    Plotting base class with parameters used by all plots

    Attributes:
        station_id: str, id of the station being plotted.  Required
        figure_title: str, the title of the plot.  Required
        figure_title_show: bool, if True, show the figure title.  Default True
        start_time_epoch: float, the epoch start time of the data.  Default 0.
        params_tfr: AudioParams, parameters for plotting audio data.  Default AudioParams()
        units_time: str, label of units for time component.  Default "s"
    """
    station_id: str
    figure_title: str
    figure_title_show: bool = True
    start_time_epoch: float = 0.
    params_tfr: AudioParams = AudioParams()
    units_time: str = "s"


@dataclass
class MeshBase:
    """
    Base class for Mesh plots.  Independent of PlotBase.

    Attributes:
        time: np.ndarray of the timestamps.  Required
        frequency: np.ndarray of the frequency data.  Required
        frequency_scaling: str, scaling for the frequency.  refer to AXIS_SCALE_VALS for options.  Default "log"
        shading: str, shading of the plot.  options: "auto", "gouraud", "flat", "nearest".  Default "auto"
        frequency_hz_ymin: optional float, minimum frequency.  Default None
        frequency_hz_ymax: optional float, maximum frequency.  Default None
        colormap: Optional str, colormap.  Default None
        units_frequency: str, units of the frequency.  Default "Hz"
    """
    time: np.ndarray
    frequency: np.ndarray
    frequency_scaling: str = "log"
    shading: str = "auto"
    frequency_hz_ymin: Optional[float] = None
    frequency_hz_ymax: Optional[float] = None
    colormap: Optional[str] = None
    units_frequency: str = "Hz"

    def __post_init__(self):
        # Autoscale to mesh frequency range
        if self.frequency_hz_ymax is None:
            self.frequency_hz_ymax = self.frequency[-1]
        if self.frequency_hz_ymin is None:
            self.frequency_hz_ymin = self.frequency[0]
        if self.frequency_scaling not in AXIS_SCALE_VALS:
            self.frequency_scaling = "log"
        if self.shading not in MESH_SHADING_VALS:
            self.shading = "auto"

    def get_shading_as_literal(self) -> Literal["auto", "gouraud", "flat", "nearest"]:
        """
        :return: Literal value of the shading
        """
        return cast(Literal, self.shading)

    def get_colormesh_params(self) -> Tuple[Optional[np.ndarray], Optional[np.ndarray], Optional[str]]:
        """
        :return: time, frequency, and shading for colormesh.  All three could be None
        """
        if self.shading in ["auto", "gouraud"]:
            return self.time, self.frequency, self.get_shading_as_literal()
        return None, None, None


def mesh_colormap_limits(
        mesh_array: np.ndarray,
        colormap_scaling: str = "auto",
        color_range: float = 16.
) -> Tuple[float, float]:
    """
    Find colormap limits for plotting

    :param mesh_array: array with mesh
    :param colormap_scaling: one of: "auto" (max/min of input mesh),
        "range" (max of input mesh - color range given) or
        "abs" (absolute max/min of input mesh).  Default "auto"
    :param color_range: range of colors.  Default is 16.0
    :return: colormap min and max values
    """
    if colormap_scaling == "auto":
        color_max = np.max(mesh_array)
        color_min = np.min(mesh_array)
    elif colormap_scaling == "range":
        color_max = np.max(mesh_array)
        color_min = color_max - color_range
    else:
        color_max = np.max(np.abs(mesh_array))
        color_min = np.min(np.abs(mesh_array))

    return color_min, color_max


@dataclass
class MeshPanel:
    """
    Panel for Mesh plots

    Attributes:
        tfr: np.ndarray of the data to plot.  Required
        colormap_scaling: str, scaling for the colormap.  options: "auto", "range", "else".  Default "auto"
        color_max: float, maximum color value.  Default 15
        color_range: float, range of color values.  Default 15
        color_min: float, minimum color value.  Default 0
        cbar_units: str, units to display for colorbar.  Default "bits"
        ytick_style: str, style for yticks.  options: "sci", "scientific", "plain".  Default "sci"
    """
    tfr: np.ndarray
    colormap_scaling: str = "auto"
    color_max: float = 15.
    color_range: float = 15.
    color_min: float = 0.
    cbar_units: str = "bits"
    ytick_style: str = "sci"

    def __post_init__(self):
        if self.colormap_scaling not in COLORMAP_SCALING_VALS:
            self.colormap_scaling = "else"
        else:
            self.set_color_min_max()
        if self.ytick_style not in YTICK_STYLE_VALS:
            self.ytick_style = "sci"

    def set_color_min_max(self):
        """
        Set color min-max if the colormap_scaling is "auto" or "range"
        """
        if self.is_auto_color_min_max():
            self.color_min, self.color_max = mesh_colormap_limits(self.tfr, self.colormap_scaling, self.color_range)

    def is_auto_color_min_max(self) -> bool:
        """
        :return: True if colormap_scaling is "auto" or "range", which auto-sets the color range
        """
        return self.colormap_scaling in ["auto", "range"]


@dataclass
class WaveformPlotBase(PlotBase):
    """
    Base class for Waveform plots.  Extends PlotBase; refer to PlotBase for additional attributes

    Attributes:
        label_panel_show: bool, if True, show the label.  Default False
        labels_fontweight: optional str, font weight of the labels.  Default None
        waveform_color: optional str, color of the waveform.  Default None
    """
    label_panel_show: bool = False
    labels_fontweight: Optional[str] = None
    waveform_color: Optional[str] = None


@dataclass
class WaveformPanel:
    """
    Panel for Waveform plots

    Attributes:
        sig: np.ndarray, the signal to plot.  Required
        time: np.ndarray, the timestamps of the data.  Required
        units: str, units of the signal.  Default "Norm"
        label: str, label for the data.  Default "(wf)"
        yscaling: str, scaling for y-axis.  options: "auto", "symmetric", "positive", "else".  Default "auto"mesh_
        ytick_style: str, style for yticks.  options: "sci", "scientific", "plain".  Default "plain"
    """
    sig: np.ndarray
    time: np.ndarray
    units: str = "Norm"
    label: str = "(wf)"
    yscaling: str = "auto"
    ytick_style: str = "plain"

    def __post_init__(self):
        if self.ytick_style not in YTICK_STYLE_VALS:
            self.ytick_style = "plain"
        if self.yscaling not in WF_Y_SCALING_VALS:
            self.yscaling = "else"

    def set_y_lims(self, axis: plt.Axes):
        """
        :param axis: the axis to update y limits for
        :return: updated axis with new y limits
        """
        if self.yscaling == "auto":
            axis.set_ylim(np.min(self.sig), np.max(self.sig))
            self.ytick_style = "plain"
        elif self.yscaling == "symmetric":
            axis.set_ylim(-np.max(np.abs(self.sig)), np.max(np.abs(self.sig)))
        elif self.yscaling == "positive":
            axis.set_ylim(0, np.max(np.abs(self.sig)))
        else:
            axis.set_ylim(-10, 10)


@dataclass
class CwPanel:
    """
    Panel for Continuous Waveform plots

    Attributes:
        sig: np.ndarray, the signal to plot.  Required
        time: np.ndarray, the timestamps of the data.  Required
        y_units: str, units of the signal.  Default "Norm"
        x_units: str, label for the time axis.  Default "s"
        title: str, title for the panel.  Default "CW"
    """
    sig: np.ndarray
    time: np.ndarray
    y_units: str = "Norm"
    x_units: str = "s"
    title: str = "CW"

    def is_no_data(self) -> bool:
        """
        :return: True if no data, False otherwise
        """
        return self.time[0] == self.time[-1]


@dataclass
class PowerPanelData:
    """
    Data to be plotted in a PowerPanel.  All values are required

    Attributes:
        sig: np.ndarray, the signal to plot.
        freq: np.ndarray, the frequencies to plot.
        linestyle: str, linestyle; all possible values:
            [‘solid’ | ‘dashed’, ‘dashdot’, ‘dotted’ | '-' | '--' | '-.' | ':' | 'None' | ' ' | ''].
        linewidth: float, linewidth.
        sig_label: str, label for the signal.
    """
    sig: np.ndarray
    freq: np.ndarray
    linestyle: str
    linewidth: float
    sig_label: str


@dataclass
class PowerPanel:
    """
    Panel for Power plots

    Attributes:
        panel_data: List of PowerPanelData objects to plot.  Required
        y_units: str, unit label for y-axis.  Default "Power/Var(signal)"
        x_units: str, unit label for x-axis.  Default "Frequency, Hz"
        title: str, title of the panel.  Default "Power"
    """
    panel_data: List[PowerPanelData]
    y_units: str = "Power/Var(signal)"
    x_units: str = "Frequency, Hz"
    title: str = "Power"


@dataclass
class CwPowerPlotBase:
    """
    Base class for plotting CW and Power plots.

    Attributes:
        params_tfr: AudioParams, parameters for plotting data.  Default AudioParams()
        figure_title_show: bool, if True, show the figure title.  Default True
    """
    params_tfr: AudioParams = AudioParams()
    figure_title_show: bool = True



import unittest
import numpy as np
import quantum_inferno.utilities.matrix as mat


class MyTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.array_1d = np.array([1, 2, 3])
        cls.array_1d_repeat_1x6 = np.array([1, 2, 3, 1, 2, 3])
        cls.array_2d_repeat_3x2 = np.array([[1, 2, 3], [1, 2, 3]])
        cls.array_3d_ones_4x4 = np.ones((3, 4, 4))

    def test_array_from_number(self):
        self.assertTrue(np.array_equal(mat.array_from_number(5, (3, 4)), np.full((3, 4), 5)))
        self.assertTrue(np.array_equal(mat.array_from_number(5, (3, 4, 4)), np.full((3, 4, 4), 5)))

    def test_n_tile_array(self):
        self.assertTrue(
            np.array_equal(mat.n_tile_array(self.array_1d, 2, mat.MatrixAxis.ROW), self.array_2d_repeat_3x2)
        )
        self.assertTrue(
            np.array_equal(mat.n_tile_array(self.array_1d, 2, mat.MatrixAxis.COLUMN), self.array_2d_repeat_3x2.T)
        )
        self.assertTrue(np.array_equal(mat.n_tile_array(self.array_1d, 0, mat.MatrixAxis.ROW), self.array_1d))
        self.assertTrue(np.array_equal(mat.n_tile_array(self.array_1d, 0, mat.MatrixAxis.COLUMN), self.array_1d))

    def test_tile_array_to_shape(self):
        self.assertTrue((mat.tile_array_to_shape(self.array_1d, (1, 6), mat.MatrixAxis.ROW) == self.array_1d).all())
        self.assertTrue(
            (mat.tile_array_to_shape(self.array_1d, (3, 2), mat.MatrixAxis.ROW) == self.array_2d_repeat_3x2).all()
        )
        self.assertTrue(
            (mat.tile_array_to_shape(self.array_1d, (2, 3), mat.MatrixAxis.COLUMN) == self.array_2d_repeat_3x2.T).all()
        )



"""
Demonstrate the synthetic benchmark signals.
"""
import matplotlib.pyplot as plt

import quantum_inferno.synth.benchmark_signals as bsig


if __name__ == "__main__":
    sig_wf, sig_t = bsig.synth_00()
    plt.figure()
    plt.plot(sig_t, sig_wf)
    plt.title("Synth 00")

    sig_wf, sig_t = bsig.synth_01()
    plt.figure()
    plt.plot(sig_t, sig_wf)
    plt.title("Synth 01")

    sig_wf, sig_t = bsig.synth_02()
    plt.figure()
    plt.plot(sig_t, sig_wf)
    plt.title("Synth 02")

    sig_wf, sig_t = bsig.synth_03()
    plt.figure()
    plt.plot(sig_t, sig_wf)
    plt.title("Synth 03")

    plt.show()



import unittest

import numpy as np

import quantum_inferno.utilities.sampling as samp


class TestSampling(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.timeseries = np.array([10, 20, 30, 40, 50, 60, 70, 80, 70, 60, 50, 40, 30, 20])

    def test_subsample(self):
        result, new_rate = samp.subsample(self.timeseries, 1.0, 3, "nth")
        self.assertEqual(len(result), 5)
        self.assertEqual(result[0], 10)
        self.assertEqual(result[-1], 30)
        self.assertAlmostEqual(new_rate, 0.33, 2)

    def test_subsample_avg(self):
        result, new_rate = samp.subsample(self.timeseries, 1.0, 3, "average")
        self.assertEqual(len(result), 4)
        self.assertEqual(result[0], 20)
        self.assertEqual(result[-1], 50)
        self.assertAlmostEqual(new_rate, 0.33, 2)

    def test_subsample_med(self):
        result, new_rate = samp.subsample(self.timeseries, 1.0, 3, "median")
        self.assertEqual(len(result), 4)
        self.assertEqual(result[0], 20)
        self.assertEqual(result[-1], 50)
        self.assertAlmostEqual(new_rate, 0.33, 2)

    def test_subsample_max(self):
        result, new_rate = samp.subsample(self.timeseries, 1.0, 3, "max")
        self.assertEqual(len(result), 4)
        self.assertEqual(result[0], 30)
        self.assertEqual(result[-1], 60)
        self.assertAlmostEqual(new_rate, 0.33, 2)

    def test_subsample_min(self):
        result, new_rate = samp.subsample(self.timeseries, 1.0, 3, "min")
        self.assertEqual(len(result), 4)
        self.assertEqual(result[0], 10)
        self.assertEqual(result[-1], 40)
        self.assertAlmostEqual(new_rate, 0.33, 2)

    def test_resample_uneven_timestamps(self):
        timestamps = np.array([10, 15, 22, 31, 41, 46, 53, 60, 75, 79, 89, 92, 93, 100])
        result, new_rate = samp.resample_uneven_timeseries(self.timeseries, timestamps, 0.2)
        self.assertEqual(len(result), 18)
        self.assertEqual(result[0], 10)
        self.assertAlmostEqual(result[-1], 27.14, 2)
        self.assertAlmostEqual(result[4], 38.89, 2)
        self.assertEqual(new_rate, 0.2)

    def test_resample_uneven_timestamps_assumed(self):
        timestamps = np.array([10, 15, 22, 31, 41, 46, 53, 60, 75, 79, 89, 92, 93, 100])
        result, new_rate = samp.resample_uneven_timeseries(self.timeseries, timestamps)
        self.assertEqual(len(result), 13)
        self.assertEqual(result[0], 10)
        self.assertAlmostEqual(result[-1], 29.89, 2)
        self.assertAlmostEqual(result[4], 46.69, 2)
        self.assertAlmostEqual(new_rate, 0.14, 2)

    def test_resample_with_sample_rate(self):
        result, new_rate = samp.resample_with_sample_rate(self.timeseries, 1.0, 0.5)
        self.assertEqual(len(result), 7)
        self.assertAlmostEqual(result[0], 12.47, 2)
        self.assertAlmostEqual(result[-1], 30.32, 2)
        self.assertAlmostEqual(new_rate, 0.5, 2)

    def test_subsample_2d(self):
        array = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]])
        result = samp.subsample_2d(array, 2, "nth")
        self.assertEqual(result.shape, (3, 2))
        self.assertEqual(result[0][0], 1)
        self.assertEqual(result[-1][-1], 11)

    def test_subsample_2d_avg(self):
        array = np.array([[1, 2, 3, 4, 5, 6, 7], [8, 9, 10, 11, 12, 13, 14], [15, 16, 17, 18, 19, 20, 21]])
        result = samp.subsample_2d(array, 3, "average")
        self.assertEqual(result.shape, (3, 2))
        self.assertEqual(result[0][0], 2)
        self.assertEqual(result[-1][-1], 19)

    def test_subsample_2d_med(self):
        array = np.array([[1, 2, 3, 4, 5, 6, 7], [8, 9, 10, 11, 12, 13, 14], [15, 16, 17, 18, 19, 20, 21]])
        result = samp.subsample_2d(array, 3, "median")
        self.assertEqual(result.shape, (3, 2))
        self.assertEqual(result[0][0], 2)
        self.assertEqual(result[-1][-1], 19)

    def test_subsample_2d_max(self):
        array = np.array([[1, 2, 3, 4, 5, 6, 7], [8, 9, 10, 11, 12, 13, 14], [15, 16, 17, 18, 19, 20, 21]])
        result = samp.subsample_2d(array, 3, "max")
        self.assertEqual(result.shape, (3, 2))
        self.assertEqual(result[0][0], 3)
        self.assertEqual(result[-1][-1], 20)

    def test_subsample_2d_min(self):
        array = np.array([[1, 2, 3, 4, 5, 6, 7], [8, 9, 10, 11, 12, 13, 14], [15, 16, 17, 18, 19, 20, 21]])
        result = samp.subsample_2d(array, 3, "min")
        self.assertEqual(result.shape, (3, 2))
        self.assertEqual(result[0][0], 1)
        self.assertEqual(result[-1][-1], 18)

    def test_decimate_timeseries(self):
        result = samp.decimate_timeseries(np.concatenate([self.timeseries, self.timeseries]), 3)
        self.assertEqual(len(result), 10)
        self.assertAlmostEqual(result[0], 9.80, 2)
        self.assertAlmostEqual(result[-1], 19.35, 2)

    def test_decimate_timeseries_2(self):
        result = samp.decimate_timeseries(np.concatenate([self.timeseries, self.timeseries]), 2)
        self.assertEqual(len(result), 14)
        self.assertAlmostEqual(result[0], 9.94, 2)
        self.assertAlmostEqual(result[-1], 29.28, 2)

    def test_decimate_collection(self):
        repeated_timeseries = np.concatenate([self.timeseries, self.timeseries])
        result = samp.decimate_timeseries_collection(np.array([repeated_timeseries]), 3)
        self.assertEqual(len(result), 1)
        self.assertEqual(len(result[0]), 10)
        self.assertAlmostEqual(result[0][0], 9.80, 2)
        self.assertAlmostEqual(result[0][-1], 19.35, 2)



"""
This module contains functionalized versions of the quantum inferno plot templates used by examples.
These functions are intended to run with the full set of parameters for each chart, without the
user knowing the underlying classes that group the values.
"""
from typing import List, Union
import numpy as np
import matplotlib.pyplot as plt

from quantum_inferno.plot_templates import figure_attributes as fa
from quantum_inferno.plot_templates import plot_base as plt_base
from quantum_inferno.plot_templates import plot_templates as plt_tpl


def plot_wf_mesh_vert_example(
        station_id: str,
        wf_panel_a_sig: np.ndarray,
        wf_panel_a_time: np.ndarray,
        mesh_time: np.ndarray,
        mesh_frequency: np.ndarray,
        mesh_panel_b_tfr: np.ndarray,
        params_tfr=plt_base.AudioParams(),
        frequency_scaling: str = "log",
        mesh_shading: str = "auto",
        wf_panel_a_yscaling: str = "auto",
        wf_panel_a_ytick_style: str = "plain",
        mesh_panel_b_ytick_style: str = "sci",
        mesh_panel_b_colormap_scaling: str = "auto",
        mesh_panel_b_color_max: float = 15,
        mesh_panel_b_color_range: float = 15,
        mesh_panel_b_color_min: float = 0,
        start_time_epoch: float = 0,
        frequency_hz_ymin: float = None,
        frequency_hz_ymax: float = None,
        mesh_colormap: str = None,
        units_time: str = "s",
        units_frequency: str = "Hz",
        wf_panel_a_units: str = "Norm",
        mesh_panel_b_cbar_units: str = "bits",
        figure_title: str = "Time-Frequency Representation",
        figure_title_show: bool = True,
):
    """
    Plot 2 vertical panels - mesh (top panel) and signal waveform (bottom panel)

    :param wf_panel_a_yscaling: 'auto', 'symmetric', 'positive'
    :param station_id: name of station
    :param wf_panel_a_sig: array with signal waveform for bottom panel
    :param wf_panel_a_time: array with signal timestamps for bottom panel
    :param mesh_time: array with mesh time
    :param mesh_frequency: array with mesh frequencies
    :param mesh_panel_b_tfr: array with mesh tfr data for mesh plot (top panel)
    :param params_tfr: Display parameters for tfr. Check AudioParams().
    :param frequency_scaling: "log" or "linear". Default is "log"
    :param mesh_shading: type of mesh shading, one of "auto", "gouraud" or "else". Default is "auto"
    :param mesh_panel_b_colormap_scaling: color scaling for mesh plot (top panel). One of: "auto", "range" or "else"
        (use inputs given in mesh_panel_b_color_max, mesh_panel_b_color_range, mesh_panel_b_color_min).
        Default is "auto"
    :param mesh_panel_b_color_max: maximum value for color scaling for mesh plot (top panel). Default is 15.0
    :param mesh_panel_b_color_range:range between maximum and minimum values in color scaling for scatter plot
        (top panel). Default is 15.0
    :param mesh_panel_b_color_min: minimum value for color scaling for mesh plot (top panel). Default is 0.0
    :param start_time_epoch: start time in epoch UTC. Default is 0.0
    :param frequency_hz_ymin: minimum frequency for y-axis
    :param frequency_hz_ymax: maximum frequency for y-axis
    :param mesh_colormap: a Matplotlib Colormap instance or registered colormap name. If None, inherits style sheet spec
    :param units_time: units of time. Default is "s"
    :param units_frequency: units of frequency. Default is "Hz"
    :param wf_panel_a_units: units of waveform plot (bottom panel). Default is "Norm"
    :param wf_panel_a_ytick_style: 'plain' or 'sci'. Default is "plain"
    :param mesh_panel_b_ytick_style: 'plain' or 'sci'. Default is "sci"
    :param mesh_panel_b_cbar_units: units of colorbar for mesh plot (top panel). Default is "bits"
    :param figure_title: title of figure. Default is "Time-Frequency Representation"
    :param figure_title_show: show title if True. Default is True
    :return: plot
    """
    wf_base = plt_base.WaveformPlotBase(station_id=station_id,
                                        figure_title=figure_title,
                                        figure_title_show=figure_title_show,
                                        start_time_epoch=start_time_epoch,
                                        params_tfr=params_tfr,
                                        units_time=units_time
                                        )
    mesh_base = plt_base.MeshBase(time=mesh_time, frequency=mesh_frequency,
                                  frequency_scaling=frequency_scaling, shading=mesh_shading,
                                  frequency_hz_ymin=frequency_hz_ymin,
                                  frequency_hz_ymax=frequency_hz_ymax,
                                  colormap=mesh_colormap,
                                  units_frequency=units_frequency)
    # build panels
    wf_panel = plt_base.WaveformPanel(sig=wf_panel_a_sig, time=wf_panel_a_time, units=wf_panel_a_units, label="(wf)",
                                      yscaling=wf_panel_a_yscaling, ytick_style=wf_panel_a_ytick_style)
    mesh_panel = plt_base.MeshPanel(tfr=mesh_panel_b_tfr,
                                    colormap_scaling=mesh_panel_b_colormap_scaling,
                                    color_max=mesh_panel_b_color_max,
                                    color_range=mesh_panel_b_color_range, color_min=mesh_panel_b_color_min,
                                    cbar_units=mesh_panel_b_cbar_units, ytick_style=mesh_panel_b_ytick_style)
    fig = plt_tpl.plot_mesh_wf_vert(
        mesh_base=mesh_base,
        mesh_panel=mesh_panel,
        wf_base=wf_base,
        wf_panel=wf_panel,
        sanitize_times=True)

    return fig


def plot_wf_mesh_mesh_vert_example(
        station_id: str,
        wf_panel_a_sig: np.ndarray,
        wf_panel_a_time: np.ndarray,
        mesh_time: np.ndarray,
        mesh_frequency: np.ndarray,
        mesh_panel_b_tfr: np.ndarray,
        mesh_panel_c_tfr: np.ndarray,
        params_tfr=plt_base.AudioParams(fa.AspectRatioType(3)),
        wf_panel_a_yscaling: str = "auto",
        wf_panel_a_ytick_style: str = "plain",
        mesh_panel_b_ytick_style: str = "sci",
        mesh_panel_c_ytick_style: str = "sci",
        frequency_scaling: str = "log",
        mesh_shading: str = "auto",
        mesh_panel_b_colormap_scaling: str = "auto",
        mesh_panel_b_color_max: float = 15,
        mesh_panel_b_color_range: float = 15,
        mesh_panel_b_color_min: float = 0,
        mesh_panel_c_colormap_scaling: str = "auto",
        mesh_panel_c_color_max: float = 15,
        mesh_panel_c_color_range: float = 15,
        mesh_panel_c_color_min: float = 0,
        start_time_epoch: float = 0,
        frequency_hz_ymin: float = None,
        frequency_hz_ymax: float = None,
        mesh_colormap: str = None,
        units_time: str = "s",
        units_frequency: str = "Hz",
        wf_panel_a_units: str = "Norm",
        mesh_panel_b_cbar_units: str = "bits",
        mesh_panel_c_cbar_units: str = "bits",
        figure_title: str = "Time-Frequency Representation",
        figure_title_show: bool = True,
):
    """
    Plot 3 vertical panels - mesh (top panel), mesh (middle panel) and signal waveform (bottom panel)

    :param mesh_panel_b_ytick_style: y-tick style for the middle mesh panel
    :param wf_panel_a_ytick_style: y-tick style for the waveform panel
    :param wf_panel_a_yscaling: y-scaling for the waveform panel
    :param mesh_panel_c_ytick_style: y-tick style for the top mesh panel
    :param station_id: name of station
    :param wf_panel_a_sig: array with signal waveform for bottom panel
    :param wf_panel_a_time: array with signal timestamps for bottom panel
    :param mesh_time: array with mesh time
    :param mesh_frequency: array with mesh frequencies
    :param mesh_panel_b_tfr: array with mesh tfr data for mesh plot (middle panel)
    :param mesh_panel_c_tfr: array with mesh tfr data for mesh plot (top panel)
    :param params_tfr: parameters for tfr. Check AudioParams().
    :param frequency_scaling: "log" or "linear". Default is "log"
    :param mesh_shading: type of mesh shading, one of "auto", "gouraud" or "else". Default is "auto"
     :param mesh_panel_b_colormap_scaling: color scaling for mesh plot (middle panel). One of: "auto", "range" or "else"
        (use inputs given in mesh_panel_b_color_max, mesh_panel_b_color_range, mesh_panel_b_color_min).
        Default is "auto"
    :param mesh_panel_b_color_max: maximum value for color scaling for mesh plot (middle panel). Default is 15.0
    :param mesh_panel_b_color_range: range between maximum and minimum values in color scaling for mesh plot
        (middle panel). Default is 15.0
    :param mesh_panel_b_color_min: minimum value for color scaling for mesh plot (middle panel). Default is 0.0
    :param mesh_panel_c_colormap_scaling: color scaling for mesh plot (top panel). One of: "auto", "range" or "else"
        (use inputs given in mesh_panel_c_color_max, mesh_panel_c_color_range, mesh_panel_c_color_min).
        Default is "auto"
    :param mesh_panel_c_color_max: maximum value for color scaling for mesh plot (top panel). Default is 15.0
    :param mesh_panel_c_color_range:range between maximum and minimum values in color scaling for scatter plot
        (top panel). Default is 15.0
    :param mesh_panel_c_color_min: minimum value for color scaling for mesh plot (top panel). Default is 0.0
    :param start_time_epoch: start time in epoch UTC. Default is 0.0
    :param frequency_hz_ymin: minimum frequency for y-axis
    :param frequency_hz_ymax: maximum frequency for y-axis
    :param mesh_colormap: a Matplotlib Colormap instance or registered colormap name. Default is "inferno"
    :param units_time: units of time. Default is "s"
    :param units_frequency: units of frequency. Default is "Hz"
    :param wf_panel_a_units: units of waveform plot (bottom panel). Default is "Norm"
    :param mesh_panel_b_cbar_units: units of colorbar for mesh plot (middle panel). Default is "bits"
    :param mesh_panel_c_cbar_units: units of colorbar for mesh plot (top panel). Default is "bits"
    :param figure_title: title of figure. Default is "Time-Frequency Representation"
    :param figure_title_show: show title if True. Default is True
    :return: plot
    """
    plot_base = plt_base.WaveformPlotBase(station_id=station_id,
                                          figure_title=figure_title,
                                          figure_title_show=figure_title_show,
                                          start_time_epoch=start_time_epoch,
                                          params_tfr=params_tfr,
                                          units_time=units_time
                                          )
    mesh_base = plt_base.MeshBase(time=mesh_time, frequency=mesh_frequency,
                                  frequency_scaling=frequency_scaling, shading=mesh_shading,
                                  frequency_hz_ymin=frequency_hz_ymin,
                                  frequency_hz_ymax=frequency_hz_ymax,
                                  colormap=mesh_colormap,
                                  units_frequency=units_frequency)
    # build panels
    wf_panel = plt_base.WaveformPanel(sig=wf_panel_a_sig, time=wf_panel_a_time, units=wf_panel_a_units, label="(wf)",
                                      yscaling=wf_panel_a_yscaling, ytick_style=wf_panel_a_ytick_style)
    mesh_panel_b = plt_base.MeshPanel(tfr=mesh_panel_b_tfr,
                                      colormap_scaling=mesh_panel_b_colormap_scaling,
                                      color_max=mesh_panel_b_color_max,
                                      color_range=mesh_panel_b_color_range, color_min=mesh_panel_b_color_min,
                                      cbar_units=mesh_panel_b_cbar_units, ytick_style=mesh_panel_b_ytick_style)
    mesh_panel_c = plt_base.MeshPanel(tfr=mesh_panel_c_tfr,
                                      colormap_scaling=mesh_panel_c_colormap_scaling,
                                      color_max=mesh_panel_c_color_max,
                                      color_range=mesh_panel_c_color_range, color_min=mesh_panel_c_color_min,
                                      cbar_units=mesh_panel_c_cbar_units, ytick_style=mesh_panel_c_ytick_style)
    fig = plt_tpl.plot_n_mesh_wf_vert(mesh_base, [mesh_panel_c, mesh_panel_b], plot_base, wf_panel)

    return fig


def plot_cw_and_power(
        cw_panel_sig: np.ndarray,
        power_panel_sigs: List[np.ndarray],
        cw_panel_time: np.ndarray,
        power_panel_freqs: List[np.ndarray],
        power_panel_ls: List[str] = None,
        power_panel_lw: List[int] = None,
        power_panel_sig_labels: List[str] = None,
        cw_panel_units: str = "Norm",
        power_panel_y_units: str = "Power/Var(signal)",
        power_panel_x_units: str = "Frequency, Hz",
        params_tfr=fa.AudioParams(),
        units_time: str = "s",
        cw_panel_title: str = "CW",
        power_panel_title: str = "Power",
        figure_title_show: bool = True,
) -> Union[plt.Figure, None]:
    """
    Template for CW and power plots from intro set examples

    :param cw_panel_sig: CW signal waveform
    :param power_panel_sigs: list of power signal arrays
    :param cw_panel_time: CW time series
    :param power_panel_freqs: list of power frequency arrays
    :param power_panel_ls: list of line styles for power signals
    :param power_panel_lw: list of line widths for power signals
    :param power_panel_sig_labels: list of labels for power signals
    :param cw_panel_units: units of the CW signal waveform
    :param power_panel_y_units: y-axis units of the power signals
    :param power_panel_x_units: x-axis units of the power signals
    :param params_tfr: parameters for tfr; see AudioParams() in figure_attributes.py
    :param units_time: x-axis units for the CW panel
    :param cw_panel_title: title of the CW panel
    :param power_panel_title: title of the power panel
    :param figure_title_show: show panel titles if True; default is True
    :return: the figure
    """
    cw_panel = plt_base.CwPanel(cw_panel_sig, cw_panel_time, cw_panel_units, units_time, cw_panel_title)
    power_panel_data = []
    for f in range(len(power_panel_sigs)):
        power_panel_data.append(plt_base.PowerPanelData(power_panel_sigs[f], power_panel_freqs[f], power_panel_ls[f],
                                                        power_panel_lw[f], power_panel_sig_labels[f]))
    power_panel = plt_base.PowerPanel(power_panel_data, power_panel_y_units, power_panel_x_units, power_panel_title)
    cw_pow_base = plt_base.CwPowerPlotBase(params_tfr, figure_title_show)
    return plt_tpl.plot_cw_and_power(cw_panel, power_panel, cw_pow_base)



"""
Methods for mathematical operations.
(Can't be named "math" because it's a built-in module)
"""

from scipy.integrate import cumulative_trapezoid
import numpy as np

FILL_LOCATIONS = ["start", "end"]
FILL_TYPES = ["zero", "nan", "mean", "median", "min", "max", "tail", "head"]
ROUNDING_TYPES = ["floor", "ceil", "round", "ceil_power_of_two", "floor_power_of_two"]
OUTPUT_TYPES = ["log2", "points", "pow2"]


# cumulative trapezoidal integration
def integrate_with_cumtrapz_timestamps_s(
    timestamps_s: np.ndarray, timeseries: np.ndarray, initial_value: float = 0
) -> np.ndarray:
    """
    Cumulative trapezoid integration using scipy.integrate.cumulative_trapezoid

    :param timestamps_s: timestamps in seconds
    :param timeseries: sensor waveform
    :param initial_value: initial value of the integral
    :return: integrated waveform
    """
    return cumulative_trapezoid(y=timeseries, x=timestamps_s, initial=initial_value)


def integrate_with_cumtrapz_sample_rate_hz(
    sample_rate_hz: float, timeseries: np.ndarray, initial_value: float = 0
) -> np.ndarray:
    """
    Cumulative trapezoid integration using scipy.integrate.cumulative_trapezoid

    :param sample_rate_hz: sample rate in Hz
    :param timeseries: sensor waveform
    :param initial_value: initial value of the integral
    :return: integrated waveform
    """
    return cumulative_trapezoid(y=timeseries, dx=1 / sample_rate_hz, initial=initial_value)


def derivative_with_gradient_timestamps_s(timestamps_s: np.ndarray, timeseries: np.ndarray) -> np.ndarray:
    """
    Derivative using numpy.gradient

    :param timestamps_s: timestamps in seconds
    :param timeseries: sensor waveform
    :return: derivative waveform
    """
    return np.gradient(timeseries, timestamps_s)


def derivative_with_gradient_sample_rate_hz(sample_rate_hz: float, timeseries: np.ndarray) -> np.ndarray:
    """
    Derivative using numpy.gradient

    :param sample_rate_hz: sample rate in Hz
    :param timeseries: sensor waveform
    :return: derivative waveform
    """
    return np.gradient(timeseries, 1 / sample_rate_hz)


def get_fill_from_filling_method(array_1d: np.ndarray, fill_type: str) -> float:
    """
    Returns the fill value based on the fill type

    Refer to FILL_TYPES for available options

    :param array_1d: 1D array with data to be filled
    :param fill_type: The fill type
    :return: The fill value
    """
    if len(np.shape(array_1d)) != 1:  # check if array_1d is a 1D array
        raise ValueError(f"array_1d has shape {np.shape(array_1d)} but should be a 1D array")

    if fill_type not in FILL_TYPES:
        raise ValueError(f"Invalid fill type {fill_type}, must be one of {FILL_TYPES}")
    elif fill_type == "zero":
        return 0
    elif fill_type == "nan":
        return np.nan
    elif fill_type == "mean":
        return np.mean(array_1d)  # check in place to insure a float is returned
    elif fill_type == "median":
        return np.median(array_1d)  # check in place to insure a float is returned
    elif fill_type == "min":
        return np.min(array_1d)
    elif fill_type == "max":
        return np.max(array_1d)
    elif fill_type == "tail":
        return array_1d[-1]
    elif fill_type == "head":
        return array_1d[0]


def append_fill(array_1d: np.ndarray, fill_value: float, fill_loc: str) -> np.ndarray:
    """
    Append fill value to the array based on the fill location

    Refer to FILL_LOCATIONS for available options

    :param array_1d: 1D array with data
    :param fill_value: fill value
    :param fill_loc: fill location
    :return: array with fill value appended
    """
    if fill_loc not in FILL_LOCATIONS:
        raise ValueError(f"Invalid fill location {fill_loc}, must be one of {FILL_LOCATIONS}")
    if fill_loc == "start":
        return np.insert(array_1d, 0, fill_value)
    elif fill_loc == "end":
        return np.append(array_1d, fill_value)


def derivative_with_difference_timestamps_s(
    timestamps_s: np.ndarray, timeseries: np.ndarray, fill_type: str = "zero", fill_loc: str = "end"
) -> np.ndarray:
    """
    Derivative using numpy.diff with fill options to return the same length as the input.

    Refer to FILL_TYPES for available options

    Refer to FILL_LOCATIONS for available options

    :param timestamps_s: timestamps in seconds
    :param timeseries: sensor waveform
    :param fill_type: fill type.  Default "zero"
    :param fill_loc: fill location.  Default "end"
    :return: derivative waveform with the same length as the input
    """
    derivative = np.diff(timeseries) / np.diff(timestamps_s)
    fill_value = get_fill_from_filling_method(derivative, fill_type)
    return append_fill(derivative, fill_value, fill_loc)


def derivative_with_difference_sample_rate_hz(
    sample_rate_hz: float, timeseries: np.ndarray, fill_type: str = "zero", fill_loc: str = "end"
) -> np.ndarray:
    """
    Derivative using numpy.diff with fill options to return the same length as the input

    Refer to FILL_TYPES for available options

    Refer to FILL_LOCATIONS for available options

    :param sample_rate_hz: sample rate in Hz
    :param timeseries: sensor waveform
    :param fill_type: fill type.  Default "zero"
    :param fill_loc: fill location.  Default "end"
    :return: derivative waveform with the same length as the input
    """
    derivative = np.diff(timeseries) * sample_rate_hz
    fill_value = get_fill_from_filling_method(derivative, fill_type)
    return append_fill(derivative, fill_value, fill_loc)


def round_value(value: float, rounding_type: str = "round") -> int:
    """
    Round value based on the rounding method for positive or negative floats.
    For rounding type "round", if the decimal is halfway between two integers, it will round to the nearest even integer

    Refer to ROUNDING_TYPES for available options

    :param value: value to be rounded
    :param rounding_type: method of rounding.  Default "round"
    :return: rounded value
    """
    if rounding_type not in ROUNDING_TYPES:
        raise ValueError(f"Invalid rounding type {rounding_type}, must be one of {ROUNDING_TYPES}")
    elif rounding_type == "floor":
        return int(np.floor(value))
    elif rounding_type == "ceil":
        return int(np.ceil(value))
    elif rounding_type == "round":
        return int(np.round(value))
    elif rounding_type == "ceil_power_of_two":
        return 2 ** int(np.ceil(np.log2(value)))
    elif rounding_type == "floor_power_of_two":
        return 2 ** int(np.floor(np.log2(value)))
    else:
        raise ValueError("Invalid rounding type")


def get_num_points(sample_rate_hz: float, duration_s: float, rounding_type: str, output_unit: str) -> int:
    """
    Get number of points in a waveform based on the sample rate and duration and round based on the rounding method.
    For rounding type "round", if the decimal is halfway between two integers, it will round to the nearest even integer

    :param sample_rate_hz: sample rate in Hz
    :param duration_s: duration in seconds
    :param rounding_type: rounding type
    :param output_unit: output unit (points, log2, or pow2)
    :return: number of points
    """
    if output_unit not in OUTPUT_TYPES:
        raise ValueError(f"Invalid output unit {output_unit}, must be one of {OUTPUT_TYPES}")
    elif output_unit == "points":
        return round_value(sample_rate_hz * duration_s, rounding_type)
    elif output_unit == "log2":
        return round_value(np.log2(sample_rate_hz * duration_s), rounding_type)
    elif output_unit == "pow2":
        return round_value(2 ** (sample_rate_hz * duration_s), rounding_type)



"""
Utilities related to exporting data.
"""

import os


def check_dir(dir_name: str) -> None:
    """
    Check if directory exists, and if not, create it.

    :param dir_name: name of directory to check
    """
    existing_dir: bool = os.path.isdir(dir_name)
    if not existing_dir:
        os.makedirs(dir_name)



"""
Base class for plotting figures.  Sets most common attributes for all plots.
"""
import enum

import numpy as np


class FigureAttributes:
    """
    This is the most basic parent class -- sets the plot canvas as well as figure handling functions. This is where
    figure size, font style and sizes, line weights and colors, etc. are established. All subsequent plot classes
    will inherit these attributes, overriding them if necessary.

    Attributes:
        fig_scale: float, scale factor for the figure.  Default 2.0
        fig_dpi: int, dots per inch for the figure.  Default 300
        ratio: np.array, 2d array of figure width, height ratio.  Default [640, 400]
        font_size_1st_level: int, scale for fontsize level 1 (titles, axes labels...).  Default 5
        font_size_2nd_level: int, scale for fontsize level 2 (legend labels, ticks...).  Default 4
        line_color: str, color for line in plot.  Default 'k' (black)
        line_style: str, style of line in plot.  Default '-' (solid line)
        fig_aspect_ratio: np.array, figure aspect ratio.  Default fig_scale * ratio
        fig_face_color: str, face color of the figure.  Default "w" (white)
        fig_edge_color: str, edge color of the figure.  Default fig_face_color
        fig_size: np.array, size of the figure.  Default fig_aspect_ratio / fig_dpi
        font_color: str, color of the font.  Default "k" (black)
        font_weight: str, weight of the font.  Default "normal"
        line_weight: int, weight of the line.  Default fig_scale * 1
        tick_size: int, size of the ticks.  Default font_size_2nd_level
        legend_label_size: int, size of the legend labels.  Default font_size_2nd_level
        fig: figure object.  Default None
    """
    def __init__(self, fig_size_ratio=np.array([640, 400]), fontsize1_scale=5, fontsize2_scale=4,
                 line_color='k', line_style='-'):
        """
        Initialize the FigureAttributes

        :param fig_size_ratio: 2d array of figure width, height ratio.  Default [640, 400]
        :param fontsize1_scale: int, scale for fontsize level 1 (titles, axes labels...).  Default 5
        :param fontsize2_scale: int, scale for fontsize level 2 (legend labels, ticks...)  Default 4
        :param line_color: string, color for line in plot.  Default 'k' (black)
        :param line_style: string, style of line in plot.  Default '-' (solid line)
        """
        self.fig_scale = 2.0
        self.fig_dpi = 300
        self.ratio = fig_size_ratio
        self.font_size_1st_level = np.rint(self.fig_scale * fontsize1_scale)
        self.font_size_2nd_level = np.rint(self.fig_scale * fontsize2_scale)
        self.line_color = line_color
        self.line_style = line_style
        self.fig_aspect_ratio = np.rint(self.fig_scale * self.ratio)
        self.fig_face_color = "w"
        self.fig_edge_color = self.fig_face_color
        self.fig_size = self.fig_aspect_ratio / self.fig_dpi
        self.font_color = "k"
        self.font_weight = "normal"
        self.line_weight = np.rint(self.fig_scale * 1)
        self.tick_size = self.font_size_2nd_level
        self.legend_label_size = self.font_size_2nd_level
        self.fig = None


class FigureAttributesBackInBlack(FigureAttributes):
    """
    Adapts FigureAttributes to have black coloring
    Refer to FigureAttributes for attribute information.
    Notable changes to default values:
        fig_face_color: "k" (black)
        font_color: "w" (white)
    """
    def __init__(
            self, fig_size_ratio=np.array([640, 400]),
            fontsize1_scale=5, fontsize2_scale=4, line_color="w", line_style="-"
    ):
        """
        Initialize the FigureAttributes

        :param fig_size_ratio: 2d array of figure width, height ratio.  Default [640, 400]
        :param fontsize1_scale: int, scale for fontsize level 1 (titles, axes labels...).  Default 5
        :param fontsize2_scale: int, scale for fontsize level 2 (legend labels, ticks...)  Default 4
        :param line_color: string, color for line in plot.  Default 'w' (white)
        :param line_style: string, style of line in plot.  Default '-' (solid line)
        """
        super().__init__(fig_size_ratio, fontsize1_scale, fontsize2_scale, line_color, line_style)
        self.fig_face_color = "k"
        self.fig_edge_color = self.fig_face_color
        self.font_color = "w"


class AspectRatioType(enum.Enum):
    """
    Enumeration denoting aspect ratios
    """
    R640x360 = 1
    R1280x720 = 2
    R1920x1080 = 3
    R2560x1440 = 4
    R3840x2160 = 5


class FigureParameters:
    """
    Allows users to set the figure size.  Refer to AspectRatioType for available aspect ratios.
    Default aspect ratio is 3840x2160.

    Attributes:
        width: int, width in pixels of the plot
        height: int, height in pixels of the plot
        scale_factor: float, scale factor for text and other elements
        figure_size_x: int, size of the figure in x direction
        figure_size_y: int, size of the figure in y direction
        text_size: int, size of the text
    """
    def __init__(self, aspect_ratio: AspectRatioType):
        """
        :param aspect_ratio: refer to AspectRatioType class for valid options
        """
        if aspect_ratio == AspectRatioType.R640x360:
            self.width = 640
            self.height = 360
            self.scale_factor = 1.0 / 3.0
        elif aspect_ratio == AspectRatioType.R1280x720:
            self.width = 1280
            self.height = 720
            self.scale_factor = 2.0 / 3.0
        elif aspect_ratio == AspectRatioType.R1920x1080:
            self.width = 1920
            self.height = 1080
            self.scale_factor = 1.25
        elif aspect_ratio == AspectRatioType.R2560x1440:
            self.width = 2560
            self.height = 1440
            self.scale_factor = 4.0 / 3.0
        else:
            self.width = 3840
            self.height = 2160
            self.scale_factor = 2.0

        scale = self.scale_factor * self.height / 8
        self.figure_size_x = int(self.width / scale)
        self.figure_size_y = int(8. / self.scale_factor)
        self.text_size = int(16.0 / self.scale_factor)


# Set Aspect Ratio
class AudioParams(FigureParameters):
    """
    Audio plot figure parameters.  Refer to FigureParameters class for valid options.
    Uses 1920X1080 as default for FigureParameters

    Attributes:
        fill_gaps: bool, if True, gaps in audio are filled.  Default True
    """
    def __init__(self, aspect_ratio: AspectRatioType = AspectRatioType.R1920x1080, fill_gaps: bool = True):
        super().__init__(aspect_ratio)
        self.fill_gaps: bool = fill_gaps



import unittest

import numpy as np

from quantum_inferno.synth.benchmark_signals import well_tempered_tone
from quantum_inferno.utilities import picker


class TestPicker(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.signal, cls.timestamps, _, cls.sample_rate, cls.freq_center, cls.resolution = well_tempered_tone()
        cls.two_peak_signal = np.array([0, 1, 2, 3, 4, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -4, -3, -2, -1])

    def test_find_sample_rate_hz_from_timestamps(self):
        timestamps = np.array([10, 20, 30, 40, 50])
        sample_rate = picker.find_sample_rate_hz_from_timestamps(timestamps, "ms")
        self.assertEqual(sample_rate, 100.0)

    def test_scale_signal_by_extraction_type_max(self):
        scaled_signal = picker.scale_signal_by_extraction_type(self.two_peak_signal, "sigmax")
        self.assertEqual(len(scaled_signal), len(self.two_peak_signal))
        self.assertEqual(scaled_signal[5], 1.0)
        self.assertEqual(scaled_signal[15], -1.0)

    def test_scale_signal_by_extraction_type_min(self):
        scaled_signal = picker.scale_signal_by_extraction_type(self.two_peak_signal, "sigmin")
        self.assertEqual(len(scaled_signal), len(self.two_peak_signal))
        self.assertEqual(scaled_signal[5], -1.0)
        self.assertEqual(scaled_signal[15], 1.0)

    def test_scale_signal_by_extraction_type_abs(self):
        scaled_signal = picker.scale_signal_by_extraction_type(self.two_peak_signal, "sigabs")
        self.assertEqual(len(scaled_signal), len(self.two_peak_signal))
        self.assertEqual(scaled_signal[5], 1.0)
        self.assertEqual(scaled_signal[15], -1.0)

    def test_scale_signal_by_extraction_type_log2(self):
        scaled_signal = picker.scale_signal_by_extraction_type(self.two_peak_signal, "log2")
        self.assertEqual(len(scaled_signal), len(self.two_peak_signal))
        self.assertAlmostEqual(scaled_signal[5], 2.32, 2)
        self.assertAlmostEqual(scaled_signal[15], 2.32, 2)

    def test_scale_signal_by_extraction_type_log2max(self):
        scaled_signal = picker.scale_signal_by_extraction_type(self.two_peak_signal, "log2max")
        self.assertEqual(len(scaled_signal), len(self.two_peak_signal))
        self.assertEqual(scaled_signal[5], 1.0)
        self.assertEqual(scaled_signal[15], 1.0)

    def test_apply_bandpass(self):
        result = picker.apply_bandpass(self.signal, (100, 200), self.sample_rate)
        self.assertEqual(len(result), len(self.signal))
        self.assertAlmostEqual(result[100], 0.0, 2)

    def test_apply_bandpass_with_bad_bandpass(self):
        self.assertRaises(ValueError, picker.apply_bandpass, self.signal, (300, 100), self.sample_rate)
        self.assertRaises(ValueError, picker.apply_bandpass, self.signal, (100, 100), self.sample_rate)
        self.assertRaises(ValueError, picker.apply_bandpass, self.signal, (1, self.sample_rate), self.sample_rate)
        self.assertRaises(ValueError, picker.apply_bandpass, self.signal, (0, 1), self.sample_rate)

    def test_find_peaks_by_extraction_type_with_bandpass(self):
        result = picker.find_peaks_by_extraction_type_with_bandpass(self.signal, (100, 300), self.sample_rate)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], 1)
        self.assertEqual(result[1], 8190)

    def test_find_peaks_by_extraction_type(self):
        result = picker.find_peaks_by_extraction_type(self.two_peak_signal, "sigmax")
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], 5)

    def test_find_peaks_with_bits(self):
        result = picker.find_peaks_with_bits(self.signal, self.sample_rate)
        self.assertEqual(len(result), 65)
        self.assertEqual(result[0], 27)
        self.assertEqual(result[1], 128)

    def test_extract_signal_index_with_buffer(self):
        result = picker.extract_signal_index_with_buffer(self.sample_rate, 4000, 2, 2)
        self.assertEqual(result[0], 2400)
        self.assertEqual(result[-1], 5600)

    def test_extract_signal_with_buffer_seconds(self):
        result = picker.extract_signal_with_buffer_seconds(self.signal, self.sample_rate, 4000, 2, 2)
        self.assertEqual(len(result), 3200)
        self.assertAlmostEqual(result[0], 0.71, 2)
        self.assertAlmostEqual(result[-1], -0.95, 2)

    def test_find_peaks_to_comb_function(self):
        result = picker.find_peaks_to_comb_function(self.signal, [100, 200])
        self.assertEqual(len(result), 8192)
        self.assertEqual(result[0], 0)
        self.assertEqual(result[100], 1)
        self.assertEqual(result[200], 1)

        result = picker.find_peaks_to_comb_function(self.signal, np.array([100, 200]))
        self.assertEqual(len(result), 8192)
        self.assertEqual(result[0], 0)
        self.assertEqual(result[100], 1)
        self.assertEqual(result[200], 1)

        result = picker.find_peaks_to_comb_function(self.signal, 200)
        self.assertEqual(len(result), 8192)
        self.assertEqual(result[0], 0)
        self.assertEqual(result[200], 1)



"""
A set of functions to rescale data.
"""
from typing import Union
import numpy as np
from quantum_inferno.scales_dyadic import get_epsilon


DATA_SCALE_TYPE = ["amplitude", "power"]


def to_log2_with_epsilon(x: Union[np.ndarray, float, list]) -> Union[np.ndarray, float]:
    """
    Convert the absolute value of the data to log2 with epsilon added to avoid log(0) and log(<0) errors.

    :param x: data or value to rescale
    :return: rescaled data or value
    """
    return np.log2(np.abs(x) + get_epsilon())


def is_power_of_two(n: int) -> bool:
    """
    :param n: value to check
    :return True if n is positive and a power of 2, False otherwise
    """
    return n > 0 and not (n & (n - 1))


def to_decibel_with_epsilon(
    x: Union[np.ndarray, float, list], reference: float = 1.0, input_scaling: str = "amplitude"
) -> Union[np.ndarray, float]:
    """
    Convert data to decibels with epsilon added to avoid log(0) errors.

    :param x: data or value to rescale
    :param reference: reference value for the decibel scaling (default is None)
    :param input_scaling: the scaling type of the data (default is amplitude)
    :return: rescaled data or value as decibels
    """
    if input_scaling not in DATA_SCALE_TYPE:
        print("Invalid input scaling type.  Defaulting to amplitude.")
        input_scaling = "amplitude"
    scale_val = 10 if input_scaling == "power" else 20

    if reference == 0:
        raise ValueError("Reference value cannot be zero.")
    elif reference == 1:
        return scale_val * np.log10(np.abs(x) + get_epsilon())
    else:
        return scale_val * np.log10(np.abs(x) + get_epsilon()) - scale_val * np.log10(reference + get_epsilon())



import unittest

import numpy as np

import quantum_inferno.utilities.window as window


class TestWindow(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.ary = np.array([5, 20, 30, 40, 45, 50, 60, 70, 85])

    def test_get_tukey(self):
        result = window.get_tukey(self.ary)
        self.assertEqual(len(result), 9)
        self.assertAlmostEqual(result[0], 0, 2)
        self.assertAlmostEqual(result[-1], 0, 2)

    def test_get_tukey_alpha_0(self):
        result = window.get_tukey(self.ary, 0)
        self.assertEqual(len(result), 9)
        self.assertAlmostEqual(result[0], 1, 2)
        self.assertAlmostEqual(result[-1], 1, 2)

    def test_get_tukey_alpha_1(self):
        result = window.get_tukey(self.ary, 1)
        self.assertEqual(len(result), 9)
        self.assertAlmostEqual(result[0], 0, 2)
        self.assertAlmostEqual(result[-1], 0, 2)

    def test_get_tukey_alpha_2(self):
        result = window.get_tukey(self.ary, 2)
        self.assertEqual(len(result), 9)
        self.assertAlmostEqual(result[0], 0, 2)
        self.assertAlmostEqual(result[-1], 0, 2)

    def test_get_tukey_by_buffer_num(self):
        result = window.get_tukey_by_buffer_num(self.ary, 2)
        self.assertEqual(len(result), 9)
        self.assertAlmostEqual(result[0], 0, 2)
        self.assertAlmostEqual(result[-1], 0, 2)

    def test_get_tukey_by_buffer_num_alpha_0(self):
        result = window.get_tukey_by_buffer_num(self.ary, 2, 0)
        self.assertEqual(len(result), 9)
        self.assertAlmostEqual(result[0], 1, 2)
        self.assertAlmostEqual(result[-1], 1, 2)

    def test_get_tukey_by_buffer_num_alpha_1(self):
        result = window.get_tukey_by_buffer_num(self.ary, 2, 1)
        self.assertEqual(len(result), 9)
        self.assertAlmostEqual(result[0], 0, 2)
        self.assertAlmostEqual(result[-1], 0, 2)

    def test_get_tukey_by_buffer_s(self):
        result = window.get_tukey_by_buffer_s(self.ary, 4, 1)
        self.assertEqual(len(result), 9)
        self.assertAlmostEqual(result[0], 0, 2)
        self.assertAlmostEqual(result[-1], 0, 2)

    def test_get_tukey_by_buffer_s_alpha_0(self):
        result = window.get_tukey_by_buffer_s(self.ary, 4, 1, 0)
        self.assertEqual(len(result), 9)
        self.assertAlmostEqual(result[0], 1, 2)
        self.assertAlmostEqual(result[-1], 1, 2)

    def test_get_tukey_by_buffer_s_alpha_1(self):
        result = window.get_tukey_by_buffer_s(self.ary, 4, 1, 1)
        self.assertEqual(len(result), 9)
        self.assertAlmostEqual(result[0], 0, 2)
        self.assertAlmostEqual(result[-1], 0, 2)



"""
Utilities for calculating frequencies for both linear and logarithmic scales.
"""
from typing import Optional
import numpy as np


def get_linear_frequency_bins_range(
    sample_rate_hz: float, segment_length: int, start_hz: Optional[float] = None, end_hz: Optional[float] = None
) -> np.ndarray:
    """
    Get the frequency bins with given sample rate and segment length that matches output from scipy.signal.spectrogram.
    Default starts at 0 Hz and ends at Nyquist frequency.

    :param sample_rate_hz: sample rate of the signal
    :param segment_length: length of the segment
    :param start_hz: Optional start frequency in Hz.  If None, value is set to 0.  Default None
    :param end_hz: Optional end frequency in Hz.  If None, value is set to sample_rate_hz / 2.  Default None
    :return: frequency bins
    """
    # Default values
    if start_hz is None:
        start_hz = 0
    if end_hz is None:
        end_hz = sample_rate_hz / 2

    # Check values
    if start_hz < 0:
        print(f"Warning: start_hz ({start_hz}) is less than 0, setting to 0")
        start_hz = 0
    if end_hz > sample_rate_hz / 2:
        print(f"Warning: end_hz ({end_hz}) is greater than Nyquist frequency, setting to Nyquist frequency")
        end_hz = sample_rate_hz / 2
    if start_hz > end_hz:
        print(f"Warning: start_hz ({start_hz}) is greater than end_hz ({end_hz}), setting to 0 and Nyquist frequency")
        start_hz = 0
        end_hz = sample_rate_hz / 2
    if segment_length < 0:
        raise ValueError(f"segment_length ({segment_length}) is less than 0")
    if sample_rate_hz < 0:
        raise ValueError(f"sample_rate_hz ({sample_rate_hz}) is less than 0")
    if segment_length > sample_rate_hz:
        print(
            f"Warning: segment_length ({segment_length}) is greater than sample_rate_hz ({sample_rate_hz})"
            f", setting to sample_rate_hz"
        )
        segment_length = sample_rate_hz

    frequency_step = sample_rate_hz / segment_length
    full_range = np.arange(start=0, stop=sample_rate_hz / 2 + frequency_step, step=frequency_step)
    return full_range[(full_range >= start_hz) & (full_range <= end_hz)]


def get_shorttime_fft_frequency_bins(sample_rate_hz: float, segment_length: int) -> np.ndarray:
    """
    Get the frequency bins with given sample rate and segment length that matches output from ShortTimeFFT.
    Starts at 0 Hz and ends at Nyquist frequency.

    :param sample_rate_hz: sample rate of the signal
    :param segment_length: length of the segment
    :return: frequency bins
    """
    return get_linear_frequency_bins_range(sample_rate_hz, segment_length)


def get_band_numbers(
    sample_rate_hz: float,
    band_order: float,
    start_hz: float = None,
    end_hz: float = None,
    base: float = 10 ** 0.3,
    reference_frequency: float = 1,
) -> np.ndarray:
    """
    Get the band numbers with given sample rate, band order, start and end frequency.
    Default gets band order starting at 1 Hz and ends at Nyquist frequency.

    :param sample_rate_hz: sample rate of the signal
    :param band_order: band order
    :param start_hz: start frequency in Hz
    :param end_hz: end frequency in Hz
    :param base: base for the logarithmic scale (default 10 ** 0.3)
    :param reference_frequency: reference frequency for the logarithmic scale (default 1)
    :return: band numbers
    """
    # Default values
    if start_hz is None:
        start_hz = 1
    if end_hz is None:
        end_hz = sample_rate_hz / 2

    # Check values
    if sample_rate_hz < 0:
        raise ValueError(f"sample_rate_hz ({sample_rate_hz}) is less than 0")
    if band_order < 0:
        raise ValueError(f"band_order ({band_order}) is less than 0")
    if start_hz < 0:
        print(f"Warning: start_hz ({start_hz}) is less than or equal 0, setting to 1")
        start_hz = 1
    if end_hz > sample_rate_hz / 2:
        print(f"Warning: end_hz ({end_hz}) is greater than Nyquist frequency, setting to Nyquist frequency")
        end_hz = sample_rate_hz / 2
    if start_hz > end_hz:
        print(f"Warning: start_hz ({start_hz}) is greater than end_hz ({end_hz}), setting to 1 and Nyquist frequency")
        start_hz = 1
        end_hz = sample_rate_hz / 2

    # Calculate band numbers
    j_min = np.floor(band_order * np.log(start_hz / reference_frequency) / np.log(base))
    j_max = np.ceil(band_order * np.log(end_hz / reference_frequency) / np.log(base))
    return np.arange(j_min, j_max + 1)


def get_log_central_frequency_bins_range(
    sample_rate_hz: float,
    band_order: float,
    start_hz: float = None,
    end_hz: float = None,
    base: float = 10 ** 0.3,
    reference_frequency: float = 1,
) -> np.ndarray:
    """
    Get the central frequency bins with given sample rate, band order, start and end frequency.
    Default starts at 1 Hz and ends at Nyquist frequency.

    :param sample_rate_hz: sample rate of the signal
    :param band_order: band order
    :param start_hz: start frequency in Hz
    :param end_hz: end frequency in Hz
    :param base: base for the logarithmic scale (default 10 ** 0.3)
    :param reference_frequency: reference frequency for the logarithmic scale (default 1)
    :return: central frequency bins
    """
    # Get band numbers
    band_numbers = get_band_numbers(sample_rate_hz, band_order, start_hz, end_hz, base, reference_frequency)

    # Calculate central frequencies
    central_frequencies = reference_frequency * base ** (band_numbers / band_order)

    return central_frequencies


def get_log_edge_frequencies(
    sample_rate_hz: float,
    band_order: float,
    start_hz: float = None,
    end_hz: float = None,
    base: float = 10 ** 0.3,
    reference_frequency: float = 1,
) -> np.ndarray:
    """
    Get the edge frequencies with given sample rate, band order, start and end frequency.
    Default starts at 1 Hz and ends at Nyquist frequency.

    :param sample_rate_hz: sample rate of the signal
    :param band_order: band order
    :param start_hz: start frequency in Hz
    :param end_hz: end frequency in Hz
    :param base: base for the logarithmic scale (default 10 ** 0.3)
    :param reference_frequency: reference frequency for the logarithmic scale (default 1)
    :return: edge frequencies
    """
    # Get band numbers
    band_numbers = get_band_numbers(sample_rate_hz, band_order, start_hz, end_hz, base, reference_frequency)

    # Calculate edge frequencies
    edge_frequencies = reference_frequency * base ** ((band_numbers - 0.5) / band_order)
    edge_frequencies = np.append(
        edge_frequencies, reference_frequency * base ** ((band_numbers + 0.5) / band_order)[-1]
    )

    return edge_frequencies



"""
A set of functions to pick key portions of a signal.
"""

from typing import Tuple, Optional, Union

import numpy as np
from scipy import signal
from scipy.signal import butter, sosfiltfilt

from quantum_inferno.utilities.date_time import convert_time_unit
from quantum_inferno.utilities.rescaling import to_log2_with_epsilon


INPUT_SCALE_TYPE = ["amplitude", "log2"]
EXTRACTION_TYPE = ["sigmax", "sigmin", "sigabs", "log2", "log2max"]


def find_sample_rate_hz_from_timestamps(timestamps: np.ndarray, time_unit: str = "s") -> float:
    """
    Find the sample rate from timestamps in a given time unit (picoseconds to years defined in convert_time_unit())

    :param timestamps: input timestamps
    :param time_unit: time unit of the timestamps (default to "s")
    :return: sample rate in Hz
    """
    timestamps_seconds = convert_time_unit(timestamps, time_unit, "s")
    return 1.0 / np.mean(np.diff(timestamps_seconds))


def scale_signal_by_extraction_type(in_signal: np.ndarray, extraction_type: str = "sigmax") -> np.ndarray:
    """
    Normalize the signal based on the extraction type

    :param in_signal: input signal
    :param extraction_type: extraction type
    :return: normalized signal
    """
    if extraction_type not in EXTRACTION_TYPE:
        print("Invalid extraction type.  Defaulting to sigmax.")
        extraction_type = "sigmax"

    if extraction_type == "sigmax":
        return in_signal / np.nanmax(in_signal)
    elif extraction_type == "sigmin":
        return in_signal / np.nanmin(in_signal)
    elif extraction_type == "sigabs":
        return in_signal / np.nanmax(np.abs(in_signal))
    elif extraction_type == "log2":
        return to_log2_with_epsilon(in_signal)
    elif extraction_type == "log2max":
        return to_log2_with_epsilon(in_signal) / np.nanmax(to_log2_with_epsilon(in_signal))


def apply_bandpass(
    timeseries: np.ndarray, filter_band: Tuple[float, float], sample_rate_hz: float, filter_order: int = 7
) -> np.ndarray:
    """
    Apply a bandpass filter to the timeseries data.
    Apparently you need at least 46 values in the timeseries for this to work

    :param timeseries: input signal
    :param filter_band: bandpass filter band
    :param sample_rate_hz: sample rate of the signal
    :param filter_order: order of the filter
    :return: filtered signal
    """
    if filter_band[0] < 0 or filter_band[1] > sample_rate_hz / 2:
        raise ValueError(f"Invalid bandpass filter band, {filter_band}, for sample rate {sample_rate_hz}")
    if filter_band[0] >= filter_band[1]:
        raise ValueError(
            f"Invalid bandpass filter band, {filter_band}, the lower bound must be less than the upper bound"
        )
    sos = butter(filter_order, filter_band, fs=sample_rate_hz, btype="band", output="sos")
    return sosfiltfilt(sos, timeseries)


def find_peaks_by_extraction_type_with_bandpass(
    timeseries: np.ndarray,
    filter_band: Tuple[float, float],
    sample_rate_hz: float,
    filter_order: int = 7,
    extraction_type: str = "sigmax",
    height: Optional[float] = 0.7,
    *args,
) -> np.ndarray:
    """
    Find peaks in the timeseries data using a normalized bandpass filter

    :param timeseries: input signal
    :param filter_band: bandpass filter band
    :param sample_rate_hz: sample rate of the signal
    :param filter_order: order of the filter (default 7)
    :param extraction_type: extraction type (default SIGMAX)
    :param height: minimum height for the peaks (default 0.7)
    :param args: additional arguments for scipy's find_peaks
    :return: location of peaks in the timeseries data
    """
    filtered_timeseries = apply_bandpass(timeseries, filter_band, sample_rate_hz, filter_order)
    scaled_filtered_timeseries = scale_signal_by_extraction_type(filtered_timeseries, extraction_type)

    return signal.find_peaks(scaled_filtered_timeseries, height=height, *args)[0]


def find_peaks_by_extraction_type(
    timeseries: np.ndarray, extraction_type: str = "sigmax", height: Optional[float] = 0.7, *args
) -> np.ndarray:
    """
    Find peaks in the timeseries data by extraction type

    :param timeseries: input signal
    :param extraction_type: extraction type (default SIGMAX)
    :param height: minimum height for the peaks (default 0.7)
    :param args: additional arguments for scipy's find_peaks
    :return: location of peaks in the timeseries data
    """
    scaled_timeseries = scale_signal_by_extraction_type(timeseries, extraction_type)

    return signal.find_peaks(scaled_timeseries, height=height, *args)[0]


def find_peaks_with_bits(
    timeseries: np.ndarray,
    sample_rate_hz: float,
    scaling_type: str = "amplitude",
    threshold_bits: Optional[int] = 1,
    time_distance_seconds: Optional[float] = 0.1,
    *args,
) -> np.ndarray:
    """
    Find peaks in the timeseries data with a threshold in bits (originally picker_signal_finder)

    :param timeseries: time series
    :param sample_rate_hz: sample rate of the signal
    :param scaling_type: scaling type of the signal; either "amplitude" or "log2". (default "amplitude")
    :param threshold_bits: threshold in bits (default 1)
    :param time_distance_seconds: minimum time distance between peaks in seconds (default 0.1)
    :param args: additional arguments for scipy's find_peaks
    :return: location of peaks in the timeseries data
    """
    timeseries_in_bits = to_log2_with_epsilon(timeseries)

    if scaling_type == "log2":
        height = np.max(timeseries_in_bits) - threshold_bits
    else:
        height = np.max(timeseries) - 2 ** threshold_bits

    return signal.find_peaks(
        timeseries_in_bits, height=height, distance=int(time_distance_seconds * sample_rate_hz), *args
    )[0]


def extract_signal_index_with_buffer(
    sample_rate_hz: float, peak: int, intro_buffer_s: float, outro_buffer_s: float
) -> Tuple[int, int]:
    """
    Extract start and end index of the extracted signal with a buffer around the peak

    :param sample_rate_hz: sample rate of the signal
    :param peak: peak location
    :param intro_buffer_s: intro buffer in seconds
    :param outro_buffer_s: outro buffer in seconds
    :return: start and end index of the extracted signal
    """
    if intro_buffer_s < 0 or outro_buffer_s < 0:
        raise ValueError(f"Negative intro_buffer_s or outro_buffer_s, {intro_buffer_s}, {outro_buffer_s}")
    return peak - int(intro_buffer_s * sample_rate_hz), peak + int(outro_buffer_s * sample_rate_hz)


def extract_signal_with_buffer_seconds(
    timeseries: np.ndarray, sample_rate_hz: float, peak: int, intro_buffer_s: float, outro_buffer_s: float
) -> np.ndarray:
    """
    Extract a signal with a buffer in seconds around the peak

    :param timeseries: input signal
    :param sample_rate_hz: sample rate of the signal
    :param peak: peak location
    :param intro_buffer_s: intro buffer in seconds
    :param outro_buffer_s: outro buffer in seconds
    :return: extracted signal
    """
    intro_index, outro_index = extract_signal_index_with_buffer(sample_rate_hz, peak, intro_buffer_s, outro_buffer_s)

    if intro_index < 0:
        print(f"Warning: intro buffer exceeds the signal length, intro_index: {intro_index}")
        intro_index = 0
    if outro_index > len(timeseries):
        print(f"Warning: outro buffer exceeds the signal length, outro_index: {outro_index}")
        outro_index = len(timeseries)

    return timeseries[intro_index:outro_index]


def find_peaks_to_comb_function(timeseries: np.ndarray, peaks: Union[list, int, np.ndarray]) -> np.ndarray:
    """
    Returns a comb function of the same length as the timeseries with 1s at the peak locations and 0s elsewhere

    :param timeseries: input signal
    :param peaks: peak locations
    :return: a comb function with the peak locations
    """
    if isinstance(peaks, np.ndarray):
        peaks = peaks.tolist()

    comb_function = np.zeros(len(timeseries))
    comb_function[peaks] = 1
    return comb_function



import numpy as np
import matplotlib.pyplot as plt
import scipy.signal as signal

from quantum_inferno.synth import benchmark_signals
from quantum_inferno.utilities.rescaling import to_log2_with_epsilon
import quantum_inferno.plot_templates.plot_templates as pt
import quantum_inferno.plot_templates.plot_base as pb


if __name__ == "__main__":
    """
    Compute the spectrogram over sliding windows. Added taper, noise, and AA to signal.
    Added STFT Tukey window alpha > 0.
    The Welch method is equivalent to averaging the spectrogram over the columns.
    """
    EVENT_NAME = "tone test"
    frequency_tone_hz = 60
    [
        mic_sig,
        time_s,
        time_fft_nd,
        frequency_sample_rate_hz,
        frequency_center_fft_hz,
        frequency_resolution_fft_hz,
    ] = benchmark_signals.well_tempered_tone(
        frequency_center_hz=frequency_tone_hz,
        frequency_sample_rate_hz=800,
        time_duration_s=16,
        time_fft_s=1,
        use_fft_frequency=True,
        add_noise_taper_aa=True,
        output_desc=True
    )
    alpha = 0.25  # 25% Tukey (Cosine) window

    # Compute the spectrogram with the spectrum option
    frequency_spect_hz, time_spect_s, psd_spec_power = signal.spectrogram(
        x=mic_sig,
        fs=frequency_sample_rate_hz,
        window=("tukey", alpha),
        nperseg=time_fft_nd,
        noverlap=time_fft_nd // 2,
        nfft=time_fft_nd,
        detrend="constant",
        return_onesided=True,
        axis=-1,
        scaling="spectrum",
        mode="psd",
    )

    # Express in bits; revisit
    mic_spect_bits = to_log2_with_epsilon(np.sqrt(psd_spec_power))

    wf_base = pb.WaveformPlotBase(station_id="00", figure_title=f"waveform for {EVENT_NAME}", start_time_epoch=0)
    wf_panel_a = pb.WaveformPanel(mic_sig, time_s, units="Norm")
    fig = pt.plot_wf_3_vert(wf_base, wf_panel_a, wf_panel_a, wf_panel_a)

    fmin = 2 * frequency_resolution_fft_hz
    fmax = frequency_sample_rate_hz / 2  # Nyquist
    mesh_base = pb.MeshBase(time_spect_s, frequency_spect_hz, frequency_hz_ymin=fmin, frequency_hz_ymax=fmax)
    mesh_panel = pb.MeshPanel(mic_spect_bits, "range", cbar_units="bits")

    fig2 = pt.plot_n_mesh_wf_vert(mesh_base, [mesh_panel, mesh_panel], wf_base, wf_panel_a)

    plt.show()



"""
Utilities for matrix operations.
"""

import numpy as np
from enum import Enum


class MatrixAxis(Enum):
    ROW = "row"  # d1
    COLUMN = "column"  # d0


def array_from_number(number: float or int, shape: tuple) -> np.ndarray:
    """
    Create an array with repeated values from a given float.

    :param number: float or int value to repeat
    :param shape: shape of the array
    :return: 1D or 2D array
    """
    return np.full(shape, number)


def n_tile_array(array: np.ndarray, n: int, axis: MatrixAxis) -> np.ndarray:
    """
    Tile a 1D array n times by repeating values in a column wise or row wise direction.
    If row wise, the output array will have n rows and the same number of columns as the input array.
    If column wise, the output array will have n columns and the same number of rows as the input array.

    :param array: 1D array
    :param n: number of times to tile
    :param axis: direction to tile
    :return: 2D array with repeated values
    """
    if n < 1:
        print("Warning: n must be greater than 1. Returning original array.")
        return array

    if axis == MatrixAxis.ROW:
        return np.tile(array, (n, 1))
    elif axis == MatrixAxis.COLUMN:
        return np.reshape(np.tile(array, (1, n)), (-1, n), order="F")
    else:
        raise ValueError("Invalid direction. Must be either ROW or COLUMN.")


def tile_array_to_shape(array: np.ndarray, shape: tuple, axis: MatrixAxis = None) -> np.ndarray:
    """
    Tile an array with a given shape by repeating values in a column wise or row wise direction to match.
    If axis is not specified, the direction will be determined by the shape.
    If row wise, the output array will have n rows and the same number of columns as the input array.
    If column wise, the output array will have n columns and the same number of rows as the input array.

    :param array: 1D array
    :param shape: shape of the output array
    :param axis: direction to tile
    :return: 2D array with repeated values
    """
    if shape[0] == 1 or shape[1] == 1:
        print("Warning: shape must be greater than 1. Returning original array.")
        return array

    if axis is None:
        if shape[0] == shape[1] and shape[0] == array.shape[0]:
            print("If shape is square, defaulting to row direction.")
            axis = MatrixAxis.ROW
        elif shape[0] == array.shape[0]:
            axis = MatrixAxis.ROW
        elif shape[1] == array.shape[0]:
            axis = MatrixAxis.COLUMN
        elif array.ndim == 1:
            print("Input array is 1D, Defaulting to row direction.")
            axis = MatrixAxis.ROW
        else:
            raise ValueError("Invalid shape. Must be a multiple of the input array.")

    if axis == MatrixAxis.ROW and shape[0] == array.shape[0]:
        print("row")
        return np.tile(array, (shape[1], 1))
    elif axis == MatrixAxis.COLUMN and shape[1] == array.shape[0]:
        print("column")
        return np.reshape(np.tile(array, (1, shape[0])), (-1, shape[0]), order="F")
    else:
        raise ValueError("Invalid direction or shape. Must be either ROW or COLUMN and be a multiple of input array.")


# TODO: work on simplifying these 3 functions (Legacy code included in styx functions to be cleaned up)
def d0tile_x_d0d1(d0: float or np.ndarray, d0d1: np.ndarray) -> np.ndarray:
    """
    Create array of repeated values with dimensions that match those of energy array
    Useful to multiply frequency-dependent values to frequency-time matrices

    :param d0: 1D input vector, nominally column frequency/scale multipliers
    :param d0d1: 2D array, first dimension should be that same as d1
    :return: array with matching values
    """
    shape_out = d0d1.shape

    if len(shape_out) == 1:
        d0_matrix = np.tile(d0, (shape_out[0]))
    elif len(shape_out) == 2:
        d0_matrix = np.tile(d0, (shape_out[1], 1)).T
    else:
        raise TypeError(f"Cannot handle an array of shape {d0.shape}.")

    if d0_matrix.shape == d0d1.shape:
        d0_x_d0d1 = d0_matrix * d0d1
    else:
        raise TypeError(f"Cannot handle an array of shape {d0.shape}.")
    return d0_x_d0d1


def d1tile_x_d0d1(d1: float or np.ndarray, d0d1: np.ndarray) -> np.ndarray:
    """
    Create array of repeated values with dimensions that match those of energy array
    Useful to multiply time-dependent values to frequency-time matrices

    :param d1: 1D input vector, nominally row time multipliers
    :param d0d1: 2D array, second dimension should be that same as d1
    :return: array with matching values
    """
    shape_out = d0d1.shape

    if len(shape_out) == 1:
        d1_matrix = np.tile(d1, (shape_out[0]))
    elif len(shape_out) == 2:
        d1_matrix = np.tile(d1, (shape_out[0], 1))
    else:
        raise TypeError(f"Cannot handle an array of shape {d1.shape}.")

    if d1_matrix.shape == d0d1.shape:
        d1_x_d0d1 = d1_matrix * d0d1
    else:
        raise TypeError(f"Cannot handle an array of shape {d1.shape}.")
    return d1_x_d0d1


def just_tile_d1(d1_array1d_in: float or np.ndarray, d0d1_shape: tuple) -> np.ndarray:
    """
    Create array of repeated values with dimensions that match those of energy array
    Useful to multiply time-dependent values to frequency-time matrices

    :param d1_array1d_in: 1D input vector, nominally row time multipliers
    :param d0d1_shape: 2D array, second dimension should be that same as d1
    :return: array with matching values
    """
    if len(d0d1_shape) == 1:
        tiled_matrix = np.tile(d1_array1d_in, (d0d1_shape[0]))
    elif len(d0d1_shape) == 2 and d0d1_shape[1] == len(d1_array1d_in):
        tiled_matrix = np.tile(d1_array1d_in, (d0d1_shape[0], 1))
    else:
        raise TypeError(f"Cannot handle an array of shape {d1_array1d_in.shape}.")

    return tiled_matrix



"""
This module contains constants and functions that help with
physical to cyber conversion with preferred quasi-dyadic orders.
"""

import sys
from typing import List, Tuple, Union

import numpy as np

""" 
Smallest number > 0 for 64-, 32-, and 16-bit floats.  
Use to avoid division by zero or log zero singularities
"""
EPSILON64 = np.finfo(np.float64).eps
EPSILON32 = np.finfo(np.float32).eps
EPSILON16 = np.finfo(np.float16).eps

# Scale multiplier for scale bands of order N > 0.75
M_OVER_N = 0.75 * np.pi

"""
Standardized scales
"""


def get_epsilon() -> float:
    """
    Return epsilon for float64, float32, float16 by detecting current interpreter's max size
    """
    if sys.maxsize > 2 ** 32:
        return EPSILON64
    elif sys.maxsize > 2 ** 16:
        return EPSILON32
    else:
        return EPSILON16


class Slice:
    """
    Constants for slice calculations, supersedes inferno/slice
    """
    # Preferred Orders
    ORD1 = 1.0  # Octaves; repeats nearly every three decades (1mHz, 1Hz, 1kHz)
    ORD3 = 3.0  # 1/3 octaves; reduced temporal uncertainty for sharp transients and good for decade cycles
    ORD6 = 6.0  # 1/6 octaves, good compromise for time-frequency resolution
    ORD12 = 12.0  # Musical tones, continuous waves, long duration sweeps
    ORD24 = 24.0  # High resolution; long duration window, 24 bands per octave
    ORD48 = 48.0  # Ultra-high spectral resolution for blowing minds and interstellar communications
    # Constant Q Base
    G2 = 2.0  # Base two for perfect octaves and fractional octaves
    G3 = 10.0 ** 0.3  # Reconciles base2 and base10
    # Time
    T_PLANCK = 5.4e-44  # 2.**(-144)   Planck time in seconds
    T0S = 1e-42  # Universal Scale in S
    T1S = 1.0  # 1 second
    T100S = 100.0  # 1 hectosecond, IMS low band edge
    T1000S = 1000.0  # 1 kiloseconds = 1 mHz
    T1M = 60.0  # 1 minute in seconds
    T1H = T1M * 60.0  # 1 hour in seconds
    T1D = T1H * 24.0  # 1 day in seconds
    TU = 2.0 ** 58  # Estimated age of the known universe in seconds
    # Frequency
    F1HZ = 1.0  # 1 Hz
    F1KHZ = 1_000.0  # 1 kHz
    F0HZ = 1.0e42  # 1/Universal Scale
    FU = 2.0 ** -58  # 1/Estimated age of the known universe in s

    # NOMINAL SAMPLE RATES (REDVOX API M, 2022)
    FS1HZ = 1.0  # SOH
    FS10HZ = 10.0  # iOS Barometer
    FS30HZ = 30.0  # Android barometer
    FS80HZ = 80.0  # Infrasound
    FS200HZ = 200.0  # Android Magnetometer, Fast
    FS400HZ = 400.0  # Android Accelerometer and Gyroscope, Fast
    FS800HZ = 800.0  # Infrasound and Low Audio
    FS8KHZ = 8_000.0  # Speech Audio
    FS16KHZ = 16_000.0  # ML Audio
    FS48KHZ = 48_000.0  # Audio to Ultrasound


# DEFAULT CONSTANTS FOR TIME_FREQUENCY CANVAS
DEFAULT_SCALE_BASE = Slice.G3
DEFAULT_SCALE_ORDER = Slice.ORD3
DEFAULT_REF_FREQUENCY_HZ = Slice.F1HZ

# compute maximum size for FFT calculations
__SIZE_MAX = get_epsilon()
if __SIZE_MAX == EPSILON64:
    __MAX_FFT_LIMIT_BY_SYSTEM = 63
elif __SIZE_MAX == EPSILON32:
    __MAX_FFT_LIMIT_BY_SYSTEM = 31
else:
    __MAX_FFT_LIMIT_BY_SYSTEM = 15

DEFAULT_SCALE_ORDER_MIN: float = 0.75  # Garces (2022)
DEFAULT_FFT_POW2_POINTS_MAX: int = 2 ** __MAX_FFT_LIMIT_BY_SYSTEM  # Computational FFT limit, tuned to computing system
DEFAULT_FFT_POW2_POINTS_MIN: int = 2 ** 8  # For a tolerable display
DEFAULT_MESH_POW2_PIXELS: int = 2 ** 19  # Total of pixels per mesh, tune to plotting engine
DEFAULT_TIME_DISPLAY_S: float = 60.0  # Physical time to display; sets display truncation
VALID_SCALE_ORDERS: List[float] = [0.75, 1, 1.5, 3, 6, 12, 24, 48]  # list of valid scale orders


def scale_order_check(scale_order: float = DEFAULT_SCALE_ORDER, show_warning: bool = True) -> float:
    """
    Ensure no negative, complex, or unreasonably small orders are passed; override to 1/3 octave band
    Standard orders are one of: 1, 3, 6, 12, 24. If order < 0.75 it reverts to order = 3

    :param scale_order: Band order, preferably one of: 1, 3, 6, 12, 24.  Must be > 0.75 or reverts to N=0.75
    :param show_warning: if True, prints warning of invalid scale_order.  Default True
    :return: sanitized scale order
    """
    scale_order = np.abs(scale_order)  # Force to be a real, positive float
    if scale_order < DEFAULT_SCALE_ORDER_MIN:
        if show_warning:
            print(
                f"** Warning from scales_dyadic.scale_order_check:\n"
                f"N < {DEFAULT_SCALE_ORDER_MIN} specified, overriding using N = {DEFAULT_SCALE_ORDER_MIN}"
            )
        scale_order = DEFAULT_SCALE_ORDER_MIN
    return scale_order


def scale_multiplier(scale_order: float = DEFAULT_SCALE_ORDER) -> float:
    """
    :param scale_order: Band order, preferably one of: 1, 3, 6, 12, 24.  Must be > 0.75 or reverts to N=0.75
    :return: Scale multiplier for scale bands of order N > 0.75
    """
    return M_OVER_N * scale_order_check(scale_order)


def cycles_from_order(scale_order: float) -> float:
    """
    Compute the number of cycles M for a specified band order N.
    N is the quantization parameter for the constant Q wavelet filters

    :param scale_order: Band order, preferably one of: 1, 3, 6, 12, 24.  Must be > 0.75 or reverts to N=0.75
    :return: number of cycled per normalized angular frequency
    """
    return scale_multiplier(scale_order)


def order_from_cycles(cycles_per_scale: float) -> float:
    """
    Compute the number of cycles M for a specified band order N
    where N is the quantization parameter for the constant Q wavelet filters

    :param cycles_per_scale: Should be greater than or equal than one
    :return: number of cycled per normalized angular frequency
    """
    # A single cycle is the min req
    if np.abs(cycles_per_scale) < 1:
        cycles_per_scale = 1.0
    return scale_order_check(cycles_per_scale / M_OVER_N)


def base_multiplier(scale_order: float = DEFAULT_SCALE_ORDER, scale_base: float = DEFAULT_SCALE_BASE) -> float:
    """
    :param scale_order: Band order, preferably one of: 1, 3, 6, 12, 24.  Must be > 0.75 or reverts to N=0.75
    :param scale_base: scale base
    :return: Dyadic (log2) foundation for arbitrary base
    """
    return scale_order_check(scale_order) / np.log2(scale_base)


def scale_from_frequency_hz(
    scale_order: float, scale_frequency_center_hz: Union[np.ndarray, float], frequency_sample_rate_hz: float
) -> Tuple[Union[np.ndarray, float], Union[np.ndarray, float]]:
    """
    Non-dimensional scale and angular frequency for canonical Gabor/Morlet wavelet

    :param scale_order: Band order, preferably one of: 1, 3, 6, 12, 24.  Must be > 0.75 or reverts to N=0.75
    :param scale_frequency_center_hz: scale frequency in hz
    :param frequency_sample_rate_hz: sample rate in hz
    :return: scale_atom, scaled angular frequency
    """
    scale_angular_frequency = 2.0 * np.pi * scale_frequency_center_hz / frequency_sample_rate_hz
    scale_atom = cycles_from_order(scale_order) / scale_angular_frequency
    return scale_atom, scale_angular_frequency


def band_frequency_low_high(
    frequency_order_input: float,
    frequency_base_input: float,
    frequency_ref_input: float,
    frequency_low_input: float,
    frequency_high_input: float,
    frequency_sample_rate_input: float,
) -> Tuple[float, float, np.ndarray, float, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Evaluate Standard Logarithmic Interval Time Parameters: ALWAYS USE HZ

    :param frequency_order_input: Nth order
    :param frequency_base_input: G2 or G3
    :param frequency_ref_input: reference frequency
    :param frequency_low_input: the lowest frequency of interest
    :param frequency_high_input: highest frequency of interest
    :param frequency_sample_rate_input: sample rate
    :return: scale_order, scale_base, scale_band_number, reference frequency value, Algebraic center of frequencies,
             Geometric center of frequencies, frequency_start, frequency_end
    """
    scale_ref_input = 1 / frequency_ref_input
    scale_nyquist_input = 2 / frequency_sample_rate_input
    scale_low_input = 1 / frequency_high_input
    if scale_low_input < scale_nyquist_input:
        scale_low_input = scale_nyquist_input
    scale_high_input = 1 / frequency_low_input

    (
        scale_order,
        scale_base,
        scale_band_number,
        scale_ref,
        scale_center_algebraic,
        scale_center_geometric,
        scale_start,
        scale_end,
    ) = band_intervals_periods(
        frequency_order_input, frequency_base_input, scale_ref_input, scale_low_input, scale_high_input
    )
    frequency_ref = 1 / scale_ref
    frequency_center_geometric = 1 / scale_center_geometric
    frequency_end = 1 / scale_start
    frequency_start = 1 / scale_end
    frequency_center_algebraic = (frequency_end + frequency_start) / 2.0

    # Inherit the order, base, and frequency band number (negative of period band number because of the inversion)
    return (
        scale_order,
        scale_base,
        -scale_band_number,
        frequency_ref,
        frequency_center_algebraic,
        frequency_center_geometric,
        frequency_start,
        frequency_end,
    )


def band_intervals_periods(
    scale_order_input: float,
    scale_base_input: float,
    scale_ref_input: float,
    scale_low_input: float,
    scale_high_input: float,
    show_warnings: bool = True,
) -> Tuple[float, float, np.ndarray, float, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Evaluate Standard Logarithmic Interval Scale Parameters using time scales in seconds
    If scales are provided as frequency, previous computations convert to time.
    Designed to take bappsband to just below Nyquist, within a band edge.
    ALWAYS CONVERT TO SECONDS
    Last updated: 20200905

    :param scale_order_input: Band order N, for ISO3 use N = 1.0 or 3.0 or 6.0 or 12.0 or 24.0
    :param scale_base_input: reference base G; i.e. G3 = 10.**0.3 or G2 = 2.0
    :param scale_ref_input: time reference: in seconds
    :param scale_low_input: Lowest scale. If Nyquist scale, 2 * sample interval in seconds.
    :param scale_high_input: Highest scale of interest in seconds
    :param show_warnings: if True, show any warnings encountered.  Default True
    :return: scale_order, scale_base, scale_band_number, scale_ref, scale_center_algebraic,
            scale_center_geometric, scale_start, scale_end
    """
    # Initiate error handling, all inputs should be numeric, positive, and real
    # If not real and positive, make them so
    [scale_ref, scale_low, scale_high, scale_base, scale_order] = np.absolute(
        [scale_ref_input, scale_low_input, scale_high_input, scale_base_input, scale_order_input]
    )

    # Check for compliance with ISO3 and/or ANSI S1.11 and for scale_order = 1, 3, 6, 12, and 24
    if scale_base == Slice.G3 or scale_base == Slice.G2:
        pass
    elif scale_base < 1.0:
        if show_warnings:
            print("\nWARNING: Base must be greater than unity. Overriding to G = 2")
        scale_base = Slice.G2
    elif show_warnings:
        print("\nWARNING: Base is not ISO3 or ANSI S1.11 compliant")
        print(f"Continuing With Non-standard base = {scale_base}...")

    # Check for compliance with ISO3 for scale_order = 1, 3, 6, 12, and 24
    # and the two 'special' orders 0.75 and 1.5
    if scale_order in VALID_SCALE_ORDERS:
        pass
    elif scale_order < 0.75:
        if show_warnings:
            print("Order must be greater than 0.75. Overriding to Order 1")
        scale_order = 1
    elif show_warnings:
        print(f"\nWARNING: Recommend Orders {VALID_SCALE_ORDERS}")
        print(f"Continuing With Non-standard Order = {scale_order}...")

    # Compute scale edge and width parameters
    scale_edge = scale_base ** (1.0 / (2.0 * scale_order))
    scale_width = scale_edge - 1.0 / scale_edge

    if scale_low < Slice.T0S:
        scale_low = Slice.T0S / scale_edge
    if scale_high < scale_low:
        if show_warnings:
            print("\nWARNING: Upper scale must be larger than the lowest scale")
            print("Overriding to min = max/G\n")
        scale_low = scale_high / scale_base
    if scale_high == scale_low:
        if show_warnings:
            print("\nWARNING: Upper scale = lowest scale, returning closest band edges")
        scale_high *= scale_edge
        scale_low /= scale_edge

    # Max and min bands are computed relative to the center scale
    n_max = np.round(scale_order * np.log(scale_high / scale_ref) / np.log(scale_base))
    n_min = np.floor(scale_order * np.log(scale_low / scale_ref) / np.log(scale_base))

    # Evaluate min, ensure it stays below Nyquist period
    scale_center_n_min = scale_ref * np.power(scale_base, n_min / scale_order)
    if (scale_center_n_min < scale_low) or (scale_center_n_min / scale_edge < scale_low - get_epsilon()):
        n_min += 1

    # Check for band number anomalies
    if n_max < n_min:
        if show_warnings:
            print("\nSPECMOD: Insufficient bandwidth for Nth band specification")
            print(f"Minimum scaled bandwidth (scale_high - scale_low)/scale_center = {scale_width}")
            print("Correct scale High/Low input parameters")
            print("Apply one order")
        n_max = np.floor(np.log10(scale_high) / np.log10(scale_base))
        n_min = n_max - scale_order

    # Band number array for Nth octave
    scale_band_number = np.arange(n_min, n_max + 1)

    # Compute exact, evenly (log) distributed, constant Q,
    # Nth octave center and band edge frequencies
    scale_band_exponent = scale_band_number / scale_order

    scale_center_geometric = scale_ref * np.power(scale_base * np.ones(scale_band_number.shape), scale_band_exponent)
    scale_start = scale_center_geometric / scale_edge
    scale_end = scale_center_geometric * scale_edge
    # The spectrum is centered on the algebraic center scale
    scale_center_algebraic = (scale_start + scale_end) / 2.0

    return (
        scale_order,
        scale_base,
        scale_band_number,
        scale_ref,
        scale_center_algebraic,
        scale_center_geometric,
        scale_start,
        scale_end,
    )


def log_frequency_hz_from_fft_points(
    frequency_sample_hz: float,
    fft_points: int,
    scale_order: float = DEFAULT_SCALE_BASE,
    scale_ref_hz: float = DEFAULT_REF_FREQUENCY_HZ,
    scale_base: float = DEFAULT_SCALE_BASE,
) -> np.ndarray:
    """
    :param frequency_sample_hz: sample rate of frequency in Hz
    :param fft_points: number of fft points
    :param scale_order: Band order, preferably one of: 1, 3, 6, 12, 24.  Must be > 0.75 or reverts to N=0.75
    :param scale_ref_hz: reference frequency in Hz
    :param scale_base: scale base
    :return: array of scaled values
    """
    # TODO: Make function to round to to nearest power of two and perform all-around error checking for pow2
    # See log 2 functions below
    log2_ave_life_dyad = int(np.ceil(np.log2(fft_points)))
    # Framework constants
    scale_mult = scale_multiplier(scale_order)
    order_over_log2base = base_multiplier(scale_order, scale_base)

    # Dyadic transforms
    log2_scale_mult = np.log2(scale_mult)

    # Dependent on sensor sample rate
    log2_ref = np.log2(frequency_sample_hz / scale_ref_hz)

    # Shortest period, Nyquist limit
    # Shortest period, nominal 8/10 of Nyquist
    band_aa = int(np.ceil(order_over_log2base * (np.log2(2.5) - log2_ref)))
    # Longest period band
    band_max = int(np.floor(order_over_log2base * (log2_ave_life_dyad - log2_scale_mult - log2_ref)))

    # Stopped before Nyquist, before max
    # Stopped at 0.8 of Nyquist
    bands = np.arange(band_aa, band_max + 1)
    # Flip so frequency increases up to one band below Nyquist
    return np.flip(scale_ref_hz * scale_base ** (-bands / scale_order))



"""
This module constructs synthetic signals
"""

import numpy as np
import scipy.signal as signal
from typing import Optional, Tuple, Union
from quantum_inferno import scales_dyadic


# TODO: ADD Gabor grain, see also cwt and dyadic scales

def gabor_grain_frequencies(
    frequency_order_input: float,
    frequency_low_input: float,
    frequency_high_input: float,
    frequency_sample_rate_input: float,
    frequency_base_input: float = scales_dyadic.Slice.G2,
    frequency_ref_input: float = 1.0,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Frequencies for g-chirps

    :param frequency_order_input: Nth order
    :param frequency_low_input: lowest frequency of interest
    :param frequency_high_input: highest frequency of interest
    :param frequency_sample_rate_input: sample rate
    :param frequency_base_input: G2 or G3. Default is G2
    :param frequency_ref_input: reference frequency. Default is 1.0
    :return: numpy arrays of center, start and end frequencies
    """
    (
        _,
        _,
        _,
        _,
        _,
        frequency_center,
        frequency_start,
        frequency_end,
    ) = scales_dyadic.band_frequency_low_high(
        frequency_order_input,
        frequency_base_input,
        frequency_ref_input,
        frequency_low_input,
        frequency_high_input,
        frequency_sample_rate_input,
    )

    return frequency_center, frequency_start, frequency_end


def chirp_noise_16bit(
    duration_points: int = 2 ** 12,
    sample_rate_hz: float = 80.0,
    noise_std_loss_bits: float = 4.0,
    frequency_center_hz: Optional[float] = None,
) -> np.ndarray:
    """
    Construct chirp with linear frequency sweep, white noise added, anti-aliased filter applied

    :param duration_points: number of points, length of signal. Default is 2 ** 12
    :param sample_rate_hz: sample rate in Hz. Default is 80.0
    :param noise_std_loss_bits: number of bits below signal standard deviation. Default is 4.0
    :param frequency_center_hz: Optional center frequency fc in Hz.  Default None
    :return: numpy ndarray with anti-aliased chirp with white noise
    """
    if not frequency_center_hz:
        frequency_center_hz = 8.0 / (duration_points / sample_rate_hz)
    frequency_start_hz = 0.5 * frequency_center_hz
    frequency_end_hz = sample_rate_hz / 4.0

    sig_time_s = np.arange(int(duration_points)) / sample_rate_hz
    chirp_wf = signal.chirp(
        sig_time_s, frequency_start_hz, sig_time_s[-1], frequency_end_hz, method="linear", phi=0, vertex_zero=True
    )
    chirp_wf *= taper_tukey(chirp_wf, 0.25)
    chirp_white = chirp_wf + white_noise_fbits(sig=chirp_wf, std_bit_loss=noise_std_loss_bits)
    chirp_white_aa = antialias_half_nyquist(chirp_white)

    return chirp_white_aa.astype(np.float16)


def sawtooth_noise_16bit(
    duration_points: int = 2 ** 12,
    sample_rate_hz: float = 80.0,
    noise_std_loss_bits: float = 4.0,
    frequency_center_hz: Optional[float] = None,
) -> np.ndarray:
    """
    Construct an anti-aliased sawtooth waveform with white noise

    :param duration_points: number of points, length of signal. Default is 2 ** 12
    :param sample_rate_hz: sample rate in Hz. Default is 80.0
    :param noise_std_loss_bits: number of bits below signal standard deviation. Default is 4.0
    :param frequency_center_hz: Optional center frequency fc in Hz.  Default None
    :return: numpy ndarray with anti-aliased sawtooth signal with white noise
    """
    frequency_center_hz = frequency_center_hz if frequency_center_hz else 8.0 / (duration_points / sample_rate_hz)

    sig_time_s = np.arange(int(duration_points)) / sample_rate_hz
    saw_wf = signal.sawtooth((2 * np.pi * frequency_center_hz) * sig_time_s, width=0)
    saw_wf *= taper_tukey(saw_wf, 0.25)
    saw_white = saw_wf + white_noise_fbits(sig=saw_wf, std_bit_loss=noise_std_loss_bits)
    saw_white_aa = antialias_half_nyquist(saw_white)

    return saw_white_aa.astype(np.float16)


def sawtooth_doppler_noise_16bit(phase_radians: np.ndarray, noise_std_loss_bits: float = 4.0) -> np.ndarray:
    """
    Construct an anti-aliased sawtooth waveform with white noise

    :param phase_radians: time-varying phase in radians
    :param noise_std_loss_bits: number of bits below signal standard deviation. Default is 4.0
    :return: numpy ndarray with anti-aliased sawtooth signal with white noise
    """
    saw_wf = signal.sawtooth(phase_radians, width=0)
    saw_wf *= taper_tukey(saw_wf, 0.25)
    saw_white = saw_wf + white_noise_fbits(sig=saw_wf, std_bit_loss=noise_std_loss_bits)
    saw_white_aa = antialias_half_nyquist(saw_white)
    saw_white_aa.astype(np.float16)

    return saw_white_aa


def chirp_linear_in_noise(
    snr_bits: float,
    sample_rate_hz: float,
    duration_s: float,
    frequency_start_hz: float,
    frequency_end_hz: float,
    intro_s: Union[int, float],
    outro_s: Union[int, float],
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Construct chirp with linear frequency sweep, white noise added.

    :param snr_bits: number of bits below signal standard deviation
    :param sample_rate_hz: sample rate in Hz
    :param duration_s: duration of chirp in seconds
    :param frequency_start_hz: start frequency in Hz
    :param frequency_end_hz: end frequency in Hz
    :param intro_s: number of seconds before chirp
    :param outro_s: number of seconds after chirp
    :return: numpy ndarray with waveform, numpy ndarray with time in seconds
    """
    sig_time_s = np.arange(int(sample_rate_hz * duration_s)) / sample_rate_hz
    chirp_wf = signal.chirp(
        sig_time_s, frequency_start_hz, sig_time_s[-1], frequency_end_hz, method="linear", phi=0, vertex_zero=True
    )
    chirp_wf *= taper_tukey(chirp_wf, 0.25)
    sig_wf = np.concatenate(
        (np.zeros(int(intro_s * sample_rate_hz)), chirp_wf, np.zeros(int(outro_s * sample_rate_hz)))
    )
    synth_wf = sig_wf + white_noise_fbits(sig=sig_wf, std_bit_loss=snr_bits)
    return synth_wf, np.arange(len(synth_wf)) / sample_rate_hz


def white_noise_fbits(sig: np.ndarray, std_bit_loss: float) -> np.ndarray:
    """
    Compute white noise with zero mean and standard deviation that is snr_bits below the input signal

    :param sig: detrended input signal
    :param std_bit_loss: number of bits below signal standard deviation
    :return: gaussian noise with zero mean
    """
    # This is in power, or variance.  White noise, zero mean
    return np.random.normal(0, np.std(sig) / 2.0 ** std_bit_loss, size=sig.size)


def taper_tukey(sig_or_time: np.ndarray, fraction_cosine: float) -> np.ndarray:
    """
    Constructs a symmetric Tukey window with the same dimensions as a time or signal numpy array.
    fraction_cosine = 0 is a rectangular window, 1 is a Hann window

    :param sig_or_time: input signal or time
    :param fraction_cosine: fraction of the window inside the cosine tapered window, shared between the head and tail
    :return: tukey taper window amplitude
    """
    return signal.windows.tukey(M=np.size(sig_or_time), alpha=fraction_cosine, sym=True)


def antialias_half_nyquist(synth: np.ndarray, filter_order: int = 4) -> np.ndarray:
    """
    Antialiasing filter with -3dB at 1/4 of sample rate, 1/2 of Nyquist

    :param synth: array with signal data
    :param filter_order: sets decay rate
    :return: numpy array with anti-aliased signal
    """
    # Antialiasing filter with -3dB at 1/4 of sample rate, 1/2 of Nyquist
    # Signal frequencies are scaled by Nyquist
    edge_high = 0.5
    [b, a] = signal.butter(filter_order, edge_high, btype="lowpass")
    return signal.filtfilt(b, a, np.copy(synth))


def frequency_algebraic_nth(frequency_geometric: np.ndarray, band_order_nth: float) -> np.ndarray:
    """
    Compute algebraic frequencies in band order

    :param frequency_geometric: geometric frequencies
    :param band_order_nth: Nth order of constant Q bands
    :return: algebraic frequencies from band order
    """
    return frequency_geometric * (np.sqrt(1 + 1 / (8 * band_order_nth ** 2)))



"""
Methods for calculating frequency and time-frequency representations of signals.
Try to match all the defaults...
"""

from typing import Tuple
import numpy as np
from scipy import signal

from quantum_inferno.utilities.calculations import round_value

# Create dictionaries for the types to avoid having to use Literal when running the functions
scaling_type = ["magnitude", "psd", None]
padding_type = ["zeros", "edge", "even", "odd"]


# return the Short-Time Fourier Transform (STFT) object with default parameters
def get_stft_object_tukey(
    sample_rate_hz: float, tukey_alpha: float, segment_length: int, overlap_length: int, scaling: str = "magnitude"
) -> signal.ShortTimeFFT:
    """
    Return the Short-Time Fourier Transform (STFT) object with a Tukey window using ShortTimeFFT class
    Calculates the number of fft points based on the segment length using ceil_power_of_two rounding method

    :param sample_rate_hz: sample rate of the signal
    :param tukey_alpha: shape parameter of the Tukey window
    :param segment_length: length of the segment
    :param overlap_length: length of the overlap
    :param scaling: scaling of the STFT (default is magnitude, other options are 'psd' and None)
    :return: ShortTimeFFT object
    """
    # checks
    if segment_length < overlap_length:
        print(
            f"overlap length {overlap_length} must be smaller than segment length {segment_length}"
            " using half of the segment length as the overlap length"
        )
        overlap_length = segment_length // 2

    if tukey_alpha < 0 or tukey_alpha > 1:
        print(f"Warning: Tukey alpha {tukey_alpha} must be between 0 and 1, using 0.25 as the default value")
        tukey_alpha = 0.25

    if scaling not in scaling_type:
        print(f"Warning: scaling {scaling} must be one of {scaling_type}, using 'magnitude' as the default value")
        scaling = "magnitude"

    # calculate the values to be used in the ShortTimeFFT object
    tukey_window = signal.windows.tukey(segment_length, alpha=tukey_alpha)
    fft_points = round_value(segment_length, "ceil_power_of_two")
    hop_length = segment_length - overlap_length

    # create the ShortTimeFFT object
    stft_obj = signal.ShortTimeFFT(
        win=tukey_window, hop=hop_length, fs=sample_rate_hz, mfft=fft_points, fft_mode="onesided", scale_to=scaling
    )

    return stft_obj


def stft_tukey(
    timeseries: np.ndarray,
    sample_rate_hz: float or int,
    tukey_alpha: float,
    segment_length: int,
    overlap_length: int,
    scaling: str = "magnitude",
    padding: str = "zeros",
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Calculate the Short-Time Fourier Transform (STFT) of a signal with a Tukey window using ShortTimeFFT class
    Returns the time, frequency, and magnitude of the STFT similar to legacy scipy.signal.stft

    :param timeseries: input signal
    :param sample_rate_hz: sample rate of the signal
    :param tukey_alpha: shape parameter of the Tukey window
    :param segment_length: length of the segment
    :param overlap_length: length of the overlap
    :param scaling: scaling of the STFT (default is None, other options are 'magnitude' and 'psd)
    :param padding: padding method for the STFT (default is 'zeros', other options are 'edge', 'even', and 'odd')
    :return: time, frequency, and magnitude of the STFT
    """
    # check if padding is valid
    if padding not in padding_type:
        print(f"Warning: padding {padding} must be one of {padding_type}, using 'zeros' as the default value")
        padding = "zeros"

    # create the ShortTimeFFT object
    stft_obj = get_stft_object_tukey(sample_rate_hz, tukey_alpha, segment_length, overlap_length, scaling)

    # calculate the STFT with detrending
    stft_magnitude = stft_obj.stft_detrend(x=timeseries, detr="constant", padding=padding)

    # calculate the time and frequency bins
    time_bins = np.arange(start=0, stop=stft_obj.delta_t * np.shape(stft_magnitude)[1], step=stft_obj.delta_t)
    frequency_bins = stft_obj.f

    return frequency_bins, time_bins, stft_magnitude


# get inverse Short-Time Fourier Transform (iSTFT) with default parameters
def istft_tukey(
    stft_to_invert: np.ndarray,
    sample_rate_hz: float or int,
    tukey_alpha: float,
    segment_length: int,
    overlap_length: int,
    scaling: str = "magnitude",
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Calculate the inverse Short-Time Fourier Transform (iSTFT) of a signal with a Tukey window using ShortTimeFFT class

    :param stft_to_invert: The STFT to be inverted
    :param sample_rate_hz: sample rate of the signal
    :param tukey_alpha: shape parameter of the Tukey window
    :param segment_length: length of the segment
    :param overlap_length: length of the overlap
    :param scaling: scaling of the STFT (default is None, other options are 'magnitude' and 'psd)
    :return: timestamps and iSTFT of the signal
    """
    # create the ShortTimeFFT object
    stft_obj = get_stft_object_tukey(sample_rate_hz, tukey_alpha, segment_length, overlap_length, scaling)

    # The index of the last window where only half of the window contains the signal
    last_window_index = int((np.shape(stft_to_invert)[1] - 1) * stft_obj.hop)

    # return timestamps for the iSTFT that includes the full signal
    timestamps = np.arange(start=0, stop=last_window_index / sample_rate_hz, step=1 / sample_rate_hz)

    return timestamps, stft_obj.istft(stft_to_invert, k1=last_window_index)


# get the spectrogram with default parameters
def spectrogram_tukey(
    timeseries: np.ndarray,
    sample_rate_hz: float or int,
    tukey_alpha: float,
    segment_length: int,
    overlap_length: int,
    scaling: str = "magnitude",
    padding: str = "zeros",
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Calculate the Spectrogram of a signal with a Tukey window using ShortTimeFFT class
    Returns the time, frequency, and spectrogram similar to legacy scipy.signal.spectrogram

    :param timeseries: input signal
    :param sample_rate_hz: sample rate of the signal
    :param tukey_alpha: shape parameter of the Tukey window
    :param segment_length: length of the segment
    :param overlap_length: length of the overlap
    :param scaling: scaling of the spectrogram (default is 'magnitude', other options are 'psd' and None)
    :param padding: padding of the signal (default is 'zeros', other options are 'periodic' and 'constant')
    :return: time, frequency, and magnitude of the STFT
    """
    # check if padding is valid
    if padding not in padding_type:
        print(f"Warning: padding {padding} must be one of {padding_type}, using 'zeros' as the default value")
        padding = "zeros"

    # Make the ShortTimeFFT object
    stft_obj = get_stft_object_tukey(sample_rate_hz, tukey_alpha, segment_length, overlap_length, scaling)

    # Calculate the spectrogram
    spectrogram = stft_obj.spectrogram(x=timeseries, padding=padding)

    # calculate the time and frequency bins
    time_bins = np.arange(start=0, stop=stft_obj.delta_t * np.shape(spectrogram)[1], step=stft_obj.delta_t)
    frequency_bins = stft_obj.f

    return frequency_bins, time_bins, spectrogram



import unittest

import numpy as np

import quantum_inferno.synth.synthetic_signals as synth


class TestSynthetics(unittest.TestCase):
    def test_gabor_grain_frequencies(self):
        center, start, end = synth.gabor_grain_frequencies(3, .01, 100, 1)
        self.assertAlmostEqual(center[0], 0.4, 2)
        self.assertAlmostEqual(center[-1], 0.01, 2)
        self.assertAlmostEqual(start[0], 0.35, 2)
        self.assertAlmostEqual(start[-1], 0.01, 2)
        self.assertAlmostEqual(end[0], 0.45, 2)
        self.assertAlmostEqual(end[-1], 0.01, 2)

    def test_chirp_noise_16bit(self):
        sig = synth.chirp_noise_16bit()
        self.assertTrue(np.min(sig) < -1.0)
        self.assertTrue(np.max(sig) > 1.0)
        self.assertEqual(sig.size, 2**12)

    def test_sawtooth_noise_16bit(self):
        sig = synth.sawtooth_noise_16bit()
        self.assertTrue(np.min(sig) < -1.0)
        self.assertTrue(np.max(sig) > 1.0)
        self.assertEqual(sig.size, 2**12)

    def test_sawtooth_doppler_noise_16bit(self):
        phase = np.array([(n % 5) / 4 * np.pi for n in range(2**12)])
        sig = synth.sawtooth_doppler_noise_16bit(phase)
        self.assertTrue(np.min(sig) < -0.01)
        self.assertTrue(np.max(sig) > 0.01)
        self.assertEqual(sig.size, 2**12)

    def test_chirp_linear_in_noise(self):
        sig, time = synth.chirp_linear_in_noise(8, 800, 3, 0, 0, 1, 1)
        self.assertEqual(sig.size, time.size)
        self.assertEqual(sig.size, 4000)
        self.assertEqual(time[0], 0)
        self.assertEqual(time[-1], 4.99875)

    def test_white_noise_fbits(self):
        sig = synth.white_noise_fbits(np.array(range(100)), 8)
        self.assertEqual(sig.size, 100)

    def test_taper_tukey(self):
        sig = synth.taper_tukey(np.array(range(100)), 0)
        self.assertEqual(sig.size, 100)
        self.assertEqual(sig[0], sig[-1])
        self.assertEqual(sig[50], 1)
        sig = synth.taper_tukey(np.array(range(200)), 1)
        self.assertEqual(sig.size, 200)
        self.assertEqual(sig[0], sig[-1])
        self.assertAlmostEqual(sig[100], 1, 2)

    def test_antialias_half_nyquist(self):
        sig = np.array([np.sin(x) for x in range(1000)])
        half_sig = synth.antialias_half_nyquist(sig)
        self.assertEqual(half_sig.size, 1000)
        self.assertAlmostEqual(half_sig[0], -3.6e-4, 2)
        self.assertAlmostEqual(half_sig[-1], -2.57e-2, 2)
        self.assertAlmostEqual(np.mean(half_sig), -1.09e-5, 2)

    def test_frequency_algebraic_nth(self):
        sig = np.array([np.cos(x) for x in range(1000)])
        alg_freq = synth.frequency_algebraic_nth(sig, 2)
        self.assertEqual(alg_freq.size, 1000)
        self.assertAlmostEqual(alg_freq[0], 1.02, 2)
        self.assertAlmostEqual(alg_freq[-1], 1.02, 2)
        self.assertAlmostEqual(np.mean(alg_freq), 9.9e-4, 2)



import unittest

import numpy as np

import quantum_inferno.utilities.calculations as calc
import quantum_inferno.synth.benchmark_signals as sigs


class CalculationsTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.timeseries, cls.timestamps, _, cls.sample_rate, _, _ = sigs.well_tempered_tone()

    def test_integrate_with_cumtrapz_timestamps_s(self):
        result = calc.integrate_with_cumtrapz_timestamps_s(self.timestamps, self.timeseries)
        self.assertEqual(len(result), 8192)
        self.assertEqual(result[0], 0)
        self.assertAlmostEqual(result[-1], -0.0012, 4)

    def test_integrate_with_cumtrapz_sample_rate_hz(self):
        result = calc.integrate_with_cumtrapz_sample_rate_hz(self.sample_rate, self.timeseries)
        self.assertEqual(len(result), 8192)
        self.assertEqual(result[0], 0)
        self.assertAlmostEqual(result[-1], -0.0012, 4)

    def test_derivative_with_gradient_timestamps_s(self):
        result = calc.derivative_with_gradient_timestamps_s(self.timestamps, self.timeseries)
        self.assertEqual(len(result), 8192)
        self.assertAlmostEqual(result[0], -85.42, 2)
        self.assertAlmostEqual(result[3], -354.39, 2)
        self.assertAlmostEqual(result[-1], 238.02, 2)

    def test_derivative_with_gradient_sample_rate_hz(self):
        result = calc.derivative_with_gradient_sample_rate_hz(self.sample_rate, self.timeseries)
        self.assertEqual(len(result), 8192)
        self.assertAlmostEqual(result[0], -85.42, 2)
        self.assertAlmostEqual(result[3], -354.39, 2)
        self.assertAlmostEqual(result[-1], 238.02, 2)

    def test_get_fill_from_filling_method(self):
        result = calc.get_fill_from_filling_method(self.timeseries, "zero")
        self.assertEqual(result, 0)
        result = calc.get_fill_from_filling_method(self.timeseries, "nan")
        self.assertTrue(np.isnan(result))
        result = calc.get_fill_from_filling_method(self.timeseries, "mean")
        self.assertAlmostEqual(result, -1.7e-16, 2)
        result = calc.get_fill_from_filling_method(self.timeseries, "median")
        self.assertAlmostEqual(result, 1.46e-15, 2)
        result = calc.get_fill_from_filling_method(self.timeseries, "min")
        self.assertEqual(result, -1.)
        result = calc.get_fill_from_filling_method(self.timeseries, "max")
        self.assertEqual(result, 1.)
        result = calc.get_fill_from_filling_method(self.timeseries, "tail")
        self.assertAlmostEqual(result, 0.89, 2)
        result = calc.get_fill_from_filling_method(self.timeseries, "head")
        self.assertEqual(result, 1.)

    def test_get_fill_from_filling_method_invalid(self):
        with self.assertRaises(ValueError):
            calc.get_fill_from_filling_method(np.array([[1], [2]]), "zero")

    def test_append_fill_start(self):
        result = calc.append_fill(self.timeseries, 0, "start")
        self.assertEqual(len(result), 8193)
        self.assertEqual(result[0], 0)
        self.assertAlmostEqual(result[-1], .89, 2)

    def test_append_fill_end(self):
        result = calc.append_fill(self.timeseries, 0, "end")
        self.assertEqual(len(result), 8193)
        self.assertEqual(result[0], 1.)
        self.assertEqual(result[-1], 0)

    def test_derivative_with_difference_timestamps_s(self):
        result = calc.derivative_with_difference_timestamps_s(self.timestamps, self.timeseries)
        self.assertEqual(len(result), 8192)
        self.assertAlmostEqual(result[0], -85.42, 2)
        self.assertEqual(result[-1], 0.)

    def test_derivative_with_difference_sample_rate_hz(self):
        result = calc.derivative_with_difference_sample_rate_hz(10, self.timeseries)
        self.assertEqual(len(result), 8192)
        self.assertAlmostEqual(result[0], -1.07, 2)
        self.assertEqual(result[-1], 0.)

    def test_round_value(self):
        result = calc.round_value(1.5, "floor")
        self.assertEqual(result, 1)
        result = calc.round_value(1.5, "ceil")
        self.assertEqual(result, 2)
        result = calc.round_value(1.5, "round")
        self.assertEqual(result, 2)

    def test_get_num_points(self):
        result = calc.get_num_points(10.0, 10.0, "round", "points")
        self.assertEqual(result, 100)
        result = calc.get_num_points(10.0, 10.0, "round", "log2")
        self.assertEqual(result, 7)
        result = calc.get_num_points(5.0, 2.0, "round", "pow2")
        self.assertEqual(result, 1024)
